#include "chain.h"
#include "crypto.h"
#include "log.h"
#include "io.h"
#include "serial.h"
#include "util.h"
#include "string.h"
#include "config.h"
#include <stdio.h>
#include <assert.h>

//TODO: this and idx table shud be generalized

#define CHA_LOAD                        CHAIN_CHA_LOAD
#define CHA_LOAD_MUL                 1000
#define CHA_POW2                        CHAIN_CHA_POW2
#define CHA_MAX                         (1<<CHA_POW2)
#define CHA_THRESHOLD                   (CHA_MAX*CHA_LOAD/CHA_LOAD_MUL)
#define CHA_FILE_SIZE                   (CHA_MAX*CHA_SIZE)
#define CHA_FILE                        "cha"
#define CHA_TMP_FILE                    "tmp"
#define CHA_FILE_ZEROS                  CHAIN_CHA_ZEROS

//--- cha file ----------------------------------------------------------------
#define CHA_ECNT_SIZE                   sizeof(uint64_t)

#define CHA_ECNT_OFF                    ( 0                                   )
#define CHA_TABLE_OFF                   ( CHA_ECNT_OFF    + CHA_ECNT_SIZE     )

//--- cha entry ---------------------------------------------------------------
#define CHA_HASH_SIZE                   sizeof(hash_t)
#define CHA_PREV_SIZE                   sizeof(hash_t)  

#define CHA_HASH_OFF                    ( 0                                   )
#define CHA_PREV_OFF                    ( CHA_HASH_OFF    + CHA_HASH_SIZE     )

#define CHA_SIZE                        ( CHA_PREV_OFF    + CHA_PREV_SIZE     )

static void cha_path(char *buf,size_t buf_size,chain_t const *c,size_t i){
  snprintf(buf,buf_size,"%s/"CHA_FILE"%0"STR(CHA_FILE_ZEROS)"zu",c->path,i);
}

static void cha_corrupt(char const *path,bool write){
  irclog_err("chain file corrupt: '%s'",path);
  if(write) disk_unsync();
}

static FILE *cha_open(char const *path,char const *mode,bool write){
  FILE *f = file_open(path,mode);
  if(!f){
    if(write) disk_unsync();
  }
  return f;
}

static bool cha_close(FILE *f,char const *path,bool write){
  if(!file_close(f)){
    if(write) disk_unsync();
    return 0;
  }
  if(write){
    disk_written(path);
  }else{
    disk_read(path);
  }
  return 1;
}

static bool cha_seek_entry(FILE *f,size_t index,bool write){
  if(!file_seek(f,CHA_TABLE_OFF+index*CHA_SIZE,SEEK_SET)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  return 1;
}

static bool cha_read_ecnt(FILE *f,uint64_t *ecnt,bool write){
  if(!file_read(f,ecnt,CHA_ECNT_SIZE)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  *ecnt = host64le(*ecnt);
  return 1;
}

static bool cha_write_ecnt(FILE *f,uint64_t ecnt){
  ecnt = host64le(ecnt);
  if(!file_write(f,&ecnt,CHA_ECNT_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool cha_read_hash(FILE *f,hash_t hash,bool write){
  if(hash){
    if(!file_read(f,hash,CHA_HASH_SIZE)){
      file_close(f);
      if(write) disk_unsync();
      return 0;
    }
  }else{
    if(!file_seek(f,CHA_HASH_SIZE,SEEK_CUR)){
      file_close(f);
      if(write) disk_unsync();
      return 0;
    }
  }
  return 1;
}

static bool cha_write_hash(FILE *f,hash_t const hash){
  if(!file_write(f,hash,CHA_HASH_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool cha_init(chain_t *c,size_t file_idx){
  char      path[PATH_BUF];
  FILE     *f;
  hash_t    hash;

  assert(CHA_THRESHOLD <= CHA_MAX);

  cha_path(path,sizeof(path),c,file_idx);
  f = cha_open(path,"wb+",1);
  if(!f) return 0;

  if(!cha_write_ecnt(f,0)) return 0;

  hash_zero(hash);
  for(size_t x = 0;x < CHA_MAX;x++){
    if(!cha_write_hash(f,hash)) return 0;
    if(!cha_write_hash(f,hash)) return 0;
  }

  if(!cha_close(f,path,1)) return 0;
  return 1;
}

static size_t cha_hash_low32(hash_t const hash){
  uint32_t lo32;
  memcpy(&lo32,hash+HASH_SIZE-sizeof(lo32),sizeof(lo32));
  return host32le(lo32);
}

static bool cha_ins(chain_t *c,hash_t const hash,hash_t const prev){
  char      path[PATH_BUF];
  FILE     *f;
  uint64_t  ecnt;
  size_t    i_start;
  size_t    i_end;

  if(c->cha_cnt == 0 || c->cha_ecnt+1 > CHA_THRESHOLD){
    if(!cha_init(c,c->cha_cnt)) return 0;
    c->cha_cnt++;
    c->cha_ecnt = 0;
  }

  cha_path(path,sizeof(path),c,c->cha_cnt-1);
  f = cha_open(path,"rb+",1);
  if(!f) return 0;

  //TODO: chek filesize

  ecnt = c->cha_ecnt+1;
  if(!cha_write_ecnt(f,ecnt)) return 0;

  i_start  = cha_hash_low32(hash)  & CHA_MAX-1;
  i_end    = i_start-1             & CHA_MAX-1;
  for(size_t i = i_start;i != i_end;i = i+1 & CHA_MAX-1){
    hash_t i_hash;

    if(!cha_seek_entry  (f,i,1))        return 0;
    if(!cha_read_hash   (f,i_hash,1))   return 0;

    if(!hash_null(i_hash)) continue;

    if(!cha_seek_entry  (f,i,1))        return 0;
    if(!cha_write_hash  (f,hash))       return 0;
    if(!cha_write_hash  (f,prev))       return 0;
    if(!cha_close       (f,path,1))     return 0;

    c->cha_ecnt = ecnt;
    return 1;
  }

  cha_corrupt(path,1);
  return 0;
}

static bool cha_cnt(chain_t *c,size_t *cha_cnt_out){
  size_t cha_cnt = 0;
  while(1){
    char path[PATH_BUF];
    cha_path(path,sizeof(path),c,cha_cnt);
    if(!file_exists(path)) break;
    cha_cnt++;
  }
  *cha_cnt_out = cha_cnt;
  return 1;
}

static bool cha_ecnt(chain_t *c,size_t *ecnt_out){
  char      path[PATH_BUF];
  FILE     *f;
  uint64_t  ecnt;

  cha_path(path,sizeof(path),c,c->cha_cnt-1);
  if(c->cha_cnt == 0 || !file_exists(path)){
    *ecnt_out = 0;
    return 1;
  }

  f = cha_open(path,"rb+",0);
  if(!f) return 0;

  if(!cha_read_ecnt(f,&ecnt,0)) return 0;

  if(ecnt > CHA_MAX){
    cha_corrupt(path,0);
    file_close(f);
    return 0;
  }

  cha_close(f,path,0);
  *ecnt_out = ecnt;
  return 1;
}

void chain_init(chain_t *c){
  assert(CHA_THRESHOLD <= CHA_MAX);
  assert(CHA_HASH_SIZE == CHA_PREV_SIZE);

  *c->path    = 0;
  c->cha_cnt  = 0;
  c->cha_ecnt = 0;
}

void chain_path(chain_t *c,char const *path){
  strzcpy(c->path,path,sizeof(c->path));
}

bool chain_load(chain_t *c){
  if(!mkdir_p(c->path)) return 0;

  if(!cha_cnt (c,&c->cha_cnt )) return 0;
  if(!cha_ecnt(c,&c->cha_ecnt)) return 0;

#ifdef LOG_CHAIN
  log_out("chain loaded: '%s'",c->path);
#endif

  return 1;
}

bool chain_push(chain_t *c,hash_t const hash,hash_t const prev){
  if(!cha_ins(c,hash,prev)) return 0;

#ifdef LOG_CHAIN
  char hash_hex[HASH_HEX_BUF];
  hash_str_log(hash_hex,sizeof(hash_hex),hash);
  log_out("%s pushed to chain: '%s'",hash_hex,c->path);
#endif

  return 1;
}

static unsigned cha_file_lookup(chain_t      *c,
                                size_t        file_idx,
                                hash_t const  hash,
                                hash_t        prev){
  char      path[PATH_BUF];
  FILE     *f;
  size_t    i_start;
  size_t    i_end;

  cha_path(path,sizeof(path),c,file_idx);
  f = cha_open(path,"r",0);
  if(!f) return LOOKUP_IO;

  i_start  = cha_hash_low32(hash)  & CHA_MAX-1;
  i_end    = i_start-1             & CHA_MAX-1;
  for(size_t i = i_start;i != i_end;i = i+1 & CHA_MAX-1){
    hash_t    i_hash;

    if(!cha_seek_entry  (f,i,0))        return LOOKUP_IO;
    if(!cha_read_hash   (f,i_hash,0))   return LOOKUP_IO;

    if(hash_null(i_hash))               break;
    if(!hash_equal(i_hash,hash))        continue;

    if(!cha_read_hash   (f,prev,0))     return LOOKUP_IO;

    cha_close(f,path,0);
    return LOOKUP_FOUND;
  }

  cha_close(f,path,0);
  return LOOKUP_NOTFOUND;
}

unsigned chain_lookup(chain_t *c,hash_t const hash,hash_t prev){
  int ret = LOOKUP_NOTFOUND;

  if(hash_null(hash)) return LOOKUP_NOTFOUND;

  for(size_t x = c->cha_cnt-1;x < (size_t)-1;x--){
    switch(cha_file_lookup(c,x,hash,prev)){
    case LOOKUP_FOUND:
      ret = LOOKUP_FOUND;
      goto ret;

    case LOOKUP_NOTFOUND:
      break;

    case LOOKUP_IO:
      irclog_err("failed to look up hash in chain: '%s'",c->path);
      return LOOKUP_IO;
    }
  }

ret:
#ifdef LOG_CHAIN
  ;char hash_hex[HASH_HEX_BUF];
  hash_str_log(hash_hex,sizeof(hash_hex),hash);
  log_out("%s %sfound in chain: '%s'",
          hash_hex,ret == LOOKUP_FOUND ? "" : "not ",c->path);
#endif
  return ret;
}

bool chain_resolve(chain_t *c,hash_t const head,hash_t tail){
  bool      ret;
  chain_t   t;
  hash_t    hash;
  hash_t    prev;
  char     *path;
  char     *path_e;

  ret = 1;

  chain_init(&t);
  path    = t.path;
  path_e  = t.path+sizeof(t.path);
  path    = strpcpy(path,c->path,         path_e);
  path    = strpcpy(path,"/"CHA_TMP_FILE, path_e);

  if(!mkdir_p(t.path)){
    ret = 0;
    goto ret;
  }

  hash_copy(hash,head);

  while(1){
    switch(chain_lookup(c,hash,prev)){
    case LOOKUP_FOUND:
      if(!chain_push(&t,hash,prev)){
        ret = 0;
        goto ret;
      }
      hash_copy(hash,prev);
      break;

    case LOOKUP_NOTFOUND:
      goto break_while;

    case LOOKUP_IO:
      ret = 0;
      goto ret;
    }
  }
break_while:

  for(size_t x = 0;x < c->cha_cnt;x++){
    char path_del[PATH_BUF];
    cha_path(path_del,sizeof(path_del),c,x);
    if(!file_delete_zero(path_del)){
      ret = 0;
      goto ret;
    }
  }

  for(size_t x = 0;x < t.cha_cnt;x++){
    char path_src[PATH_BUF];
    char path_dst[PATH_BUF];
    cha_path(path_src,sizeof(path_src),&t,x);
    cha_path(path_dst,sizeof(path_dst),c,x);
    if(!file_rename(path_src,path_dst)){
      ret = 0;
      goto ret;
    }
  }

  c->cha_cnt  = t.cha_cnt;
  c->cha_ecnt = t.cha_ecnt;

  hash_copy(tail,hash);

ret:
  rm_rf(t.path);
  return ret;
}

