#include "pest.h"
#include "crypto.h"
#include "log.h"
#include "io.h"
#include "serial.h"
#include "def.h"
#include "config.h"
#include "state.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#define DEDUP_HEADER_SIZE     (sizeof(uint64_t)+sizeof(uint64_t)+sizeof(uint64_t))
#define DEDUP_ENTRY_SIZE      (sizeof(hash_t)+sizeof(uint64_t))
#define DEDUP_OFF             (DEDUP_HEADER_SIZE)

static bool dedup_read_header(FILE *f){
  uint64_t _f;
  uint64_t _b;
  uint64_t size;

  if(!file_read(f,&_f,sizeof(_f))){
    file_close(f);
    return 0;
  }

  if(!file_read(f,&_b,sizeof(_b))){
    file_close(f);
    return 0;
  }

  if(!file_read(f,&size,sizeof(size))){
    file_close(f);
    return 0;
  }

  dedup_b = _b == (uint64_t)-1 ? (uintptr_t)-1 : _b;
  dedup_f = host64le(_f);
  dedup_b = host64le(dedup_b);
  size    = host64le(size);

  if(size != DEDUP_MAX){
    log_err("dedup buffer size changed. plz change "
            "bak to %zu or wait 1h and delete file",size); 
    return 0;
  }

  return 1;
}

static bool dedup_write_header(FILE *f){
  uint64_t _f;
  uint64_t _b;
  uint64_t size;

  _b    = dedup_b == (uintptr_t)-1 ? (uint64_t)-1 : dedup_b;
  _f    = host64le(dedup_f);
  _b    = host64le(_b);
  size  = host64le(DEDUP_MAX);

  if(!file_write(f,&_f,sizeof(_f))){
    file_close(f);
    return 0;
  }

  if(!file_write(f,&_b,sizeof(_b))){
    file_close(f);
    return 0;
  }

  if(!file_write(f,&size,sizeof(size))){
    file_close(f);
    return 0;
  }

  return 1;
}

static bool dedup_read_entry(FILE *f,hash_t hash,uint64_t *time){
  if(!file_read(f,hash,sizeof(hash_t))){
    file_close(f);
    return 0;
  }

  if(!file_read(f,time,sizeof(uint64_t))){
    file_close(f);
    return 0;
  }
  *time = host64le(*time);

  return 1;
}

static bool dedup_write_entry(FILE *f,hash_t const hash,uint64_t time){
  if(!file_write(f,hash,sizeof(hash_t))){
    file_close(f);
    return 0;
  }

  time = host64le(time);
  if(!file_write(f,&time,sizeof(uint64_t))){
    file_close(f);
    return 0;
  }

  return 1;
}

static bool dedup_init(char const *path){
  FILE     *f;
  hash_t    hash;
  uint64_t  time;

  f = file_open(path,"wb");
  if(!f) return 0;

  if(!dedup_write_header(f)) return 0;

  hash_zero(hash);
  time = 0;
  for(size_t x = 0;x < DEDUP_MAX;x++){
    if(!dedup_write_entry(f,hash,time)){
      return 0;
    }
  }

  if(!file_close(f)){
    return 0;
  }
  disk_written(path);
  return 1;
}

static bool dedup_store(size_t idx){
  char  path[PATH_BUF];
  FILE *f;

  FMT_PATH(path,"%s",FILE_DEDUP);
  f = file_open(path,"rb+");
  if(!f){
    disk_unsync();
    return 0;
  }

  if(!dedup_write_header(f)){
    disk_unsync();
    return 0;
  }

  if(idx != (size_t)-1){
    if(!file_seek(f,DEDUP_OFF+idx*DEDUP_ENTRY_SIZE,SEEK_SET)){
      file_close(f);
      disk_unsync();
      return 0;
    }

    if(!dedup_write_entry(f,dedup_h[idx],dedup_t[idx])){
      disk_unsync();
      return 0;
    }
  }

  if(!file_close(f)){
    disk_unsync();
    return 0;
  }
  disk_written(path);
  return 1;
}

bool dedup_load(void){
  char  path[PATH_BUF];
  FILE *f;

  FMT_PATH(path,"%s",FILE_DEDUP);
  if(!file_exists(path)){
    return dedup_init(path);
  }

  //TODO: maby this descriptor shud be kept open
  f = file_open(path,"rb");
  if(!f) return 0;

  if(!dedup_read_header(f)) return 0;

  for(size_t x = 0;x < DEDUP_MAX;x++){
    if(!dedup_read_entry(f,dedup_h[x],&dedup_t[x])){
      return 0;
    }
  }

  file_close(f);
  disk_read(path);
  return 1;
}

static inline bool dedup_empty(void){
  return dedup_b+1 == 0;
}

static inline size_t dedup_cnt(void){
  if(dedup_empty()) return 0;
  intptr_t diff = (intptr_t)dedup_b-(intptr_t)dedup_f;
  if(diff < 0) diff += DEDUP_MAX;
  return ++diff;
}

static size_t dedup_lookup_idx(uint8_t const *hash){
  size_t cnt = dedup_cnt();
  for(size_t x = cnt-1;x < (size_t)-1;x--){
    size_t  idx = dedup_f+x & DEDUP_MAX-1;
    hash_t *h   = dedup_h+idx;
    if(memcmp(h,hash,sizeof(*h)) == 0) return idx;
  }
  return (size_t)-1;
}

bool dedup_lookup(uint8_t const *hash){
  size_t idx = dedup_lookup_idx(hash);
  return idx != (size_t)-1;
}

bool dedup_push(uint8_t const *hash,uint64_t time){
  uintptr_t back_next;

  back_next   = dedup_b+1;
  back_next  &= DEDUP_MAX-1;

  if(back_next == dedup_f && !dedup_empty()){
    irclog_err("dedup buffer full");
    return 0;
  }

  dedup_b           = back_next;
  dedup_t[dedup_b]  = time;
  memcpy(dedup_h+dedup_b,hash,sizeof(hash_t));
  dedup_store(dedup_b);
  return 1;
}

static void dedup_pop(void){
  dedup_f++;
  if(dedup_b+1 == dedup_f){
    dedup_f   =  0;
    dedup_b   = -1;
  }else{
    dedup_f  &= DEDUP_MAX-1;
  }
  dedup_store(-1);
}

void dedup_clean(uint64_t time){
  while(!dedup_empty()){
    if(time < dedup_t[dedup_f]+DEDUP_TIMEOUT){
      return;
    }
    dedup_pop();
  }
}

