#include "chain.h"
#include "crypto.h"
#include "log.h"
#include "io.h"
#include "util.h"
#include "string.h"

void ht_init(ht_t *ht){
  *ht->path = 0;
  ht->cnt   = 1;
  ht->last  = -1;
  hash_zero(ht->hash[0]);
}

void ht_path(ht_t *ht,char const *path){
  strzcpy(ht->path,path,sizeof(ht->path));
}

bool ht_store(ht_t *ht){
  if(!hashlist_store(ht->path,(hash_t const *)ht->hash,ht->cnt)){
    disk_unsync();
    return 0;
  }
  return 1;
}

bool ht_load(ht_t *ht){
  ht->last = -1;
  if(file_exists(ht->path)){
    return hashlist_load(ht->path,ht->hash,&ht->cnt,HEADTAIL_MAX);
  }else{
    ht->cnt = 1;
    hash_zero(ht->hash[0]);
    return ht_store(ht);
  }
}

hash_t *ht_lookup(ht_t *ht,hash_t const hash){
  for(size_t x = 0;x < ht->cnt;x++){
    hash_t *h = ht->hash+x;
    if(hash_equal(*h,hash)) return h;
  }
  return 0;
}

hash_t *ht_last(ht_t *ht){
  if(ht->last == (size_t)-1) return 0;
  return &ht->hash[ht->last];
}

bool ht_full(ht_t *ht){
  return ht->cnt >= HEADTAIL_MAX;
}

hash_t *ht_add(ht_t *ht,hash_t const hash){
  hash_t *h;

  if(ht->cnt >= HEADTAIL_MAX){
    irclog_err("max heads/tails reached: %zu",HEADTAIL_MAX); 
    return 0;
  }

  h         = &ht->hash[ht->cnt];
  ht->last  = ht->cnt;
  ht->cnt++;
  hash_copy(*h,hash);
  ht_store(ht);
  return h;
}

void ht_del(ht_t *ht,hash_t *h){
  size_t idx;

  idx = ARR_INDEXOF(ht->hash,h);
  ARR_DEL(ht->hash,ht->cnt,h);
  ht->cnt--;
  assert(ht->cnt > 0);

  if(ht->last == idx){
    //dum but alternativ is store stamps for each shit 76kb of mem
    ht->last = -1;
  }else if(ht->last > idx){
    ht->last--;
  }

  ht_store(ht);
}

void ht_set(ht_t *ht,hash_t *h,hash_t const hash){
  ht->last = ARR_INDEXOF(ht->hash,h);
  hash_copy(*h,hash);
  ht_store(ht);
}

void ht_clear(ht_t *ht,hash_t const hash){
  ht->cnt   = 1;
  ht->last  = 0;
  hash_copy(ht->hash[0],hash);
  ht_store(ht);
}

bool ht_null(ht_t *h,ht_t *t){
  return h->cnt == 1 &&
         t->cnt == 1 &&
         hash_null(h->hash[0]) &&
         hash_null(t->hash[0]);
}

size_t ht_forks(ht_t *h,ht_t *t){
  assert(h->cnt >= t->cnt);
  return h->cnt-t->cnt;
}

size_t ht_breaks(ht_t *h,ht_t *t){
  (void)h;
  assert(t->cnt >= 1);
  return t->cnt-1;
}

