#include "chain.h"
#include "crypto.h"
#include "io.h"
#include "serial.h"
#include "log.h"
#include "util.h"

#define IDX_LOAD                        MESS_IDX_LOAD
#define IDX_LOAD_MUL                 1000
#define IDX_POW2                        MESS_IDX_POW2
#define IDX_MAX                         (1<<IDX_POW2)
#define IDX_THRESHOLD                   (IDX_MAX*IDX_LOAD/IDX_LOAD_MUL)
#define IDX_FILE_SIZE                   (IDX_MAX*IDX_SIZE)
#define IDX_FILE                        "idx"
#define IDX_FILE_ZEROS                  MESS_IDX_ZEROS
#define IDX_INDEX_EMPTY                 ((uint64_t)-1)

//--- idx file ----------------------------------------------------------------
#define IDX_ECNT_SIZE                   sizeof(uint64_t)

#define IDX_ECNT_OFF                    ( 0                                   )
#define IDX_TABLE_OFF                   ( IDX_ECNT_OFF    + IDX_ECNT_SIZE     )

//--- idx entry ---------------------------------------------------------------
#define IDX_INDEX_SIZE                  sizeof(uint64_t)  
#define IDX_HASH_SIZE                   sizeof(hash_t)

#define IDX_INDEX_OFF                   ( 0                                   )
#define IDX_HASH_OFF                    ( IDX_INDEX_OFF   + IDX_INDEX_SIZE    )

#define IDX_SIZE                        ( IDX_HASH_OFF    + IDX_HASH_SIZE     )

static void idx_path(char *buf,size_t buf_size,mess_t const *c,size_t i){
  snprintf(buf,buf_size,"%s/"IDX_FILE"%0"STR(IDX_FILE_ZEROS)"zu",c->path,i);
}

static void idx_corrupt(char const *path,bool write){
  irclog_err("mess index file corrupt: '%s'",path);
  if(write) disk_unsync();
}

static FILE *idx_open(char const *path,char const *mode,bool write){
  FILE *f = file_open(path,mode);
  if(!f){
    if(write) disk_unsync();
  }
  return f;
}

static bool idx_close(FILE *f,char const *path,bool write){
  if(!file_close(f)){
    if(write) disk_unsync();
    return 0;
  }
  if(write){
    disk_written(path);
  }else{
    disk_read(path);
  }
  return 1;
}

static bool idx_seek_entry(FILE *f,size_t index,bool write){
  if(!file_seek(f,IDX_TABLE_OFF+index*IDX_SIZE,SEEK_SET)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  return 1;
}

static bool idx_read_ecnt(FILE *f,uint64_t *ecnt,bool write){
  if(!file_read(f,ecnt,IDX_ECNT_SIZE)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  *ecnt = host64le(*ecnt);
  return 1;
}

static bool idx_write_ecnt(FILE *f,uint64_t ecnt){
  ecnt = host64le(ecnt);
  if(!file_write(f,&ecnt,IDX_ECNT_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool idx_read_index(FILE *f,uint64_t *index,bool write){
  if(!file_read(f,index,IDX_INDEX_SIZE)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  *index = host64le(*index);
  return 1;
}

static bool idx_write_index(FILE *f,uint64_t index){
  index = host64le(index);
  if(!file_write(f,&index,IDX_INDEX_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool idx_read_hash(FILE *f,hash_t hash,bool write){
  if(!file_read(f,hash,IDX_HASH_SIZE)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  return 1;
}

static bool idx_write_hash(FILE *f,hash_t const hash){
  if(!file_write(f,hash,IDX_HASH_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool idx_init(mess_t *c,size_t file_idx){
  char      path[PATH_BUF];
  FILE     *f;
  hash_t    hash;

  assert(IDX_THRESHOLD <= IDX_MAX);

  idx_path(path,sizeof(path),c,file_idx);
  f = idx_open(path,"wb+",1);
  if(!f) return 0;

  if(!idx_write_ecnt(f,0)) return 0;

  hash_zero(hash);
  for(size_t x = 0;x < IDX_MAX;x++){
    if(!idx_write_index (f,IDX_INDEX_EMPTY)) return 0;
    if(!idx_write_hash  (f,hash           )) return 0;
  }

  if(!idx_close(f,path,1)) return 0;
  return 1;
}

static size_t idx_hash_low32(hash_t const hash){
  uint32_t lo32;
  memcpy(&lo32,hash+HASH_SIZE-sizeof(lo32),sizeof(lo32));
  return host32le(lo32);
}

bool idx_ins(mess_t *c,hash_t const hash,uint64_t index){
  char      path[PATH_BUF];
  FILE     *f;
  uint64_t  ecnt;
  size_t    i_start;
  size_t    i_end;

  if(c->idx_cnt == 0 || c->idx_ecnt+1 > IDX_THRESHOLD){
    if(!idx_init(c,c->idx_cnt)) return 0;
    c->idx_cnt++;
    c->idx_ecnt = 0;
  }

  idx_path(path,sizeof(path),c,c->idx_cnt-1);
  f = idx_open(path,"rb+",1);
  if(!f) return 0;

  //TODO: chek filesize

  ecnt = c->idx_ecnt+1;
  if(!idx_write_ecnt(f,ecnt)) return 0;

  i_start  = idx_hash_low32(hash)  & IDX_MAX-1;
  i_end    = i_start-1             & IDX_MAX-1;
  for(size_t i = i_start;i != i_end;i = i+1 & IDX_MAX-1){
    uint64_t i_index;

    if(!idx_seek_entry  (f,i,1))        return 0;
    if(!idx_read_index  (f,&i_index,1)) return 0;

    if(i_index != IDX_INDEX_EMPTY) continue;

    if(!idx_seek_entry  (f,i,1))        return 0;
    if(!idx_write_index (f,index))      return 0;
    if(!idx_write_hash  (f,hash))       return 0;
    if(!idx_close       (f,path,1))     return 0;

    c->idx_ecnt = ecnt;
    return 1;
  }

  idx_corrupt(path,1);
  return 0;
}

bool idx_cnt(mess_t *c,size_t *idx_cnt_out){
  size_t idx_cnt = 0;
  while(1){
    char path[PATH_BUF];
    idx_path(path,sizeof(path),c,idx_cnt);
    if(!file_exists(path)) break;
    idx_cnt++;
  }
  *idx_cnt_out = idx_cnt;
  return 1;
}

bool idx_ecnt(mess_t *c,size_t *ecnt_out){
  char      path[PATH_BUF];
  FILE     *f;
  uint64_t  ecnt;

  idx_path(path,sizeof(path),c,c->idx_cnt-1);
  if(c->idx_cnt == 0 || !file_exists(path)){
    *ecnt_out = 0;
    return 1;
  }

  f = idx_open(path,"rb+",0);
  if(!f) return 0;

  if(!idx_read_ecnt(f,&ecnt,0)) return 0;

  if(ecnt > IDX_MAX){
    idx_corrupt(path,0);
    file_close(f);
    return 0;
  }

  idx_close(f,path,0);
  *ecnt_out = ecnt;
  return 1;
}

static unsigned idx_file_lookup(mess_t       *c,
                                size_t        file_idx,
                                hash_t const  hash,
                                uint64_t     *index){
  char      path[PATH_BUF];
  FILE     *f;
  size_t    i_start;
  size_t    i_end;

  idx_path(path,sizeof(path),c,file_idx);
  f = idx_open(path,"r",0);
  if(!f) return LOOKUP_IO;

  i_start  = idx_hash_low32(hash)  & IDX_MAX-1;
  i_end    = i_start-1             & IDX_MAX-1;
  for(size_t i = i_start;i != i_end;i = i+1 & IDX_MAX-1){
    uint64_t  i_index;
    hash_t    i_hash;

    if(!idx_seek_entry  (f,i,0))        return LOOKUP_IO;
    if(!idx_read_index  (f,&i_index,0)) return LOOKUP_IO;
    if(i_index == IDX_INDEX_EMPTY) break;

    if(!idx_read_hash   (f,i_hash,0))   return LOOKUP_IO;
    if(!hash_equal(i_hash,hash)) continue;

    idx_close(f,path,0);
    *index = i_index;
    return LOOKUP_FOUND;
  }

  idx_close(f,path,0);
  return LOOKUP_NOTFOUND;
}

unsigned idx_lookup(mess_t *c,hash_t const hash,uint64_t *index){
  for(size_t x = c->idx_cnt-1;x < (size_t)-1;x--){
    switch(idx_file_lookup(c,x,hash,index)){
    case LOOKUP_FOUND:
      return LOOKUP_FOUND;

    case LOOKUP_NOTFOUND:
      break;

    case LOOKUP_IO:
      irclog_err("failed to look up index in chain: '%s'",c->path);
      return LOOKUP_IO;
    }
  }
  return LOOKUP_NOTFOUND;
}

