#include "irc.h"
#include "ircd.h"
#include "pest.h"
#include "peer.h"
#include "crypto.h"
#include "net.h"
#include "log.h"
#include "util.h"
#include "data.h"
#include "wipe.h"
#include "state.h"
#include "string.h"
#include <string.h>
#include <stdlib.h>

void irc_privmsg(char const *speaker,
                        char const *text,
                        char const *chan){

  if(!irc_ready) return;
  ircd_sendf(":%s PRIVMSG %s :%s",speaker,chan,text);
}

void irc_join(char const *speaker){
  char host[IRC_HOST_BUF];
  if(!irc_ready) return;
  fmt_peerhost(host,sizeof(host),speaker);
  ircd_sendf(":%s JOIN %s",host,irc_chan);
}

void irc_part(char const *speaker){
  char host[IRC_HOST_BUF];
  if(!irc_ready) return;
  fmt_peerhost(host,sizeof(host),speaker);
  ircd_sendf(":%s PART %s",host,irc_chan);
}

static void irc_sendmotd(void){
  ircd_sendf(":%s 375 %s :- %s Message of the Day -",
             IRC_SERV,self_handle,IRC_SERV);
  ircd_sendf(":%s 372 %s :- %s",
             IRC_SERV,self_handle,IRC_MOTD);
  ircd_sendf(":%s 376 %s :End of MOTD command.",
             IRC_SERV,self_handle);
}

static void irc_welcome(void){
  if( irc_ready  ||
     !irc_conned ||
     !irc_usered ||
     !irc_passed ||
     !irc_nicked) return;
  irc_ready = 1;

  ircd_sendf(":%s 001 %s :Welcome to pest %s!%s@%s",
             IRC_SERV,self_handle,self_handle,irc_user,irc_host);
  ircd_sendf(":%s 002 %s :Your host is %s, "
             "running version " PROG_FULLNAME,
             IRC_SERV,self_handle,IRC_SERV,PROG_VERSION);
  ircd_sendf(":%s 003 %s :This server was created at some point",
             IRC_SERV,self_handle);
  ircd_sendf(":%s 004 %s %s " PROG_FULLNAME " - -",
             IRC_SERV,self_handle,IRC_SERV,PROG_VERSION);
  irc_sendmotd();

  log_out("operator connected");
}

static void irc_names(char const *chan){
  if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) == 0){
    char  names[(HANDLE_LEN+1)*(1+PEER_MAX)+1]; //"handle "*(self+peers)\0
    char *n = names;
    char *e = names+sizeof(names);

    *n  = 0;
    n   = strpcpy(n,self_handle,e);
    for(size_t x = 0;x < peer_cnt;x++){
      peer_t *p = peer+x;
      if(!p->active) continue;
      n = strpcpy(n," ",e);
      n = strpcpy(n,p->handle[0],e);
    }

    ircd_sendf(":%s 353 %s @ %s :%s",
               IRC_SERV,self_handle,chan,names);
  }
  ircd_sendf(":%s 366 %s %s :End of NAMES list.",
             IRC_SERV,self_handle,chan);
}

static void ircop_user(ircmsg_t const *m){
  if(irc_usered) return;
  if(m->argc < 1 || m->argc > 4){
    return;
  }

  char *user = m->argv[0];
  if(strncmp(user,irc_user,IRC_USER_LEN) != 0){
    log_err("username mismatch: '%s' != '%s'",user,irc_user);
    ircd_ckill();
    return;
  }

  irc_usered = 1;
  log_out("username accepted");
  irc_welcome();
}

static void ircop_pass(ircmsg_t const *m){
  char     *pwd;
  uint8_t   pass[SHA512_SIZE];

  if(irc_passed) return;
  if(m->argc != 1){
    return;
  }

  pwd = m->argv[0];
  sha512(pass,(uint8_t *)pwd,strlen(pwd));
  if(memcmp(pass,irc_pass,SHA512_SIZE) != 0){
    log_err("password mismatch");
    ircd_ckill();
    return;
  }

  irc_passed = 1;
  log_out("password accepted");
  irc_welcome();
}

static void ircop_nick(ircmsg_t const *m){
  if(m->argc != 1){
    irc_out("usage: NICK handle");
    return;
  }

  char *handle = m->argv[0];
  if(!str_valid(handle,1,HANDLE_LEN,HANDLE_CHARSET)){
    ircd_sendf(":%s 432 %s :Erroneus handle",IRC_SERV,handle);
    irc_err("invalid handle: '%s'",handle);
    return;
  }

  if(peer_lookup_noerr(handle)){
    ircd_sendf(":%s 433 %s :Nickname is already in use",IRC_SERV,handle);
    irc_err("handle '%s' already exists in WOT",handle);
    return;
  }

  ircd_sendf(":%s!%s@%s NICK :%s",self_handle,irc_user,irc_host,handle);
  strzcpy(self_handle,handle,sizeof(self_handle));
  irc_nicked = 1;
  log_out("handle set to '%s'",self_handle);
  irc_welcome();
}

static void ircop_join(ircmsg_t const *m){
  if(m->argc < 1 || m->argc > 2){
    irc_out("usage: JOIN channel [key]");
    return;
  }

  char *chan = m->argv[0];
  if(chan[0] != '#' || strlen(chan) > IRC_CHAN_LEN){
    irc_err("invalid channel name: '%s'",chan);
    return;
  }

  if(irc_joined){
    ircd_sendf(":%s!%s@%s PART %s :",self_handle,irc_user,irc_host,irc_chan);
    log_out("operator parted %s",irc_chan);
  }
  irc_joined = 1;
  strzcpy(irc_chan,chan,sizeof(irc_chan));

  ircd_sendf(":%s!%s@%s JOIN %s",self_handle,irc_user,irc_host,irc_chan);
  irc_names(irc_chan);
  log_out("operator joined %s",chan);
}

static void ircop_part(ircmsg_t const *m){
  if(m->argc < 1 || m->argc > 2){
    irc_out("usage: PART channel [message]");
    return;
  }

  char *chan = m->argv[0];
  if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) != 0){
    ircd_sendf(":%s 442 %s :You're not on that channel",IRC_SERV,self_handle);
    return;
  }

  ircd_sendf(":%s!%s@%s PART %s :",self_handle,irc_user,irc_host,irc_chan);
  log_out("operator parted %s",irc_chan);
  irc_joined  = 0;
  *irc_chan   = 0;
}

static void ircop_privmsg(ircmsg_t const *m){
  if(m->argc != 2){
    irc_out("usage: PRIVMSG channel|handle message");
    return;
  }

  char *chan  = m->argv[0];
  char *msg   = m->argv[1];

  if(chan[0] == '#'){
    if(strncmpci(chan,irc_chan,IRC_CHAN_LEN) != 0){
      ircd_sendf(":%s 404 %s :Cannot send to channel",IRC_SERV,chan);
      return;
    }
    pest_broadcast(msg);
  }else{
    pest_direct(chan,msg);
  }
}

static void ircop_names(ircmsg_t const *m){
  if(m->argc < 1 || m->argc > 2){
    irc_out("usage: NAMES channel [target]");
    return;
  }

  char *chan = m->argv[0];
  irc_names(chan);
}

static void ircop_version(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: VERSION [target]");
    return;
  }

  ircd_sendf(":%s 351 %s %s. %s :spec " STR(PEST_VERSION),
             IRC_SERV,self_handle,PROG_FULLNAME,IRC_SERV);
}

static void ircop_motd(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: MOTD [target]");
    return;
  }

  irc_sendmotd();
}

static void ircop_ping(ircmsg_t const *m){
  if(m->argc > 2){
    irc_out("usage: PING [server1 [server2]]");
    return;
  }

  ircd_sendf(":%s PONG %s :%s",IRC_SERV,IRC_SERV,IRC_SERV);
}

static void ircop_quit(ircmsg_t const *m){
  if(m->argc > 1){
    irc_out("usage: QUIT [message]");
    return;
  }

  (void)m;
  log_out("operator disconnected");
  ircd_ckill();
}

#define IRC_OP(N)           { #N, 0, ircop_##N },
#define IRC_OP_INIT(N)      { #N, 1, ircop_##N },
#define IRC_OP_IGNORE(N)    { #N, 0, 0         },

typedef void (*ircmsg_f)(ircmsg_t const *);

typedef struct ircop{
  char const       *cmd;
  bool              init;
  ircmsg_f          func;
}ircop_t;

static ircop_t const irc_op[] = {
  IRC_OP_INIT   (user)
  IRC_OP_INIT   (pass)
  IRC_OP_INIT   (nick)
  IRC_OP        (join)
  IRC_OP        (part)
  IRC_OP        (privmsg)
  IRC_OP        (names)
  IRC_OP        (version)
  IRC_OP        (motd)
  IRC_OP        (ping)
  IRC_OP_IGNORE (cap)
  IRC_OP_IGNORE (mode)
  IRC_OP_IGNORE (notice)
  IRC_OP_IGNORE (who)
  IRC_OP_IGNORE (whois)
  IRC_OP_IGNORE (whowas)
  IRC_OP        (quit)
};

void irc_ev(unsigned e,ircmsg_t const *msg){
  switch(e){
    char cmd[IRCD_CMD_BUF];

  case IRCD_CONNECT:
    ip_str(irc_host,sizeof(irc_host),ircd_getip());
    irc_conned = 1;
    break;

  case IRCD_DISCONNECT:
    *irc_host   = 0;
    irc_conned  = 0;
    irc_usered  = 0;
    irc_passed  = 0;
    irc_nicked  = 0;
    irc_joined  = 0;
    irc_ready   = 0;
    break;

  case IRCD_MESSAGE:
    for(size_t x = 0;x < ARR_CNT(irc_op);x++){
      ircop_t const *op = irc_op+x;
      if(strncmpci(msg->cmd,op->cmd,IRCD_CMD_BUF) != 0) continue;
      if(op->init || irc_ready){
        if(op->func) op->func(msg);
      }
      return;
    }

    strzcpy(cmd,msg->cmd,sizeof(cmd));
    str_toupper(cmd);
    irc_err("unknown/unhandled IRC command received: %s",cmd);
    break;

  default:
    assert(0);
  }
}

