#include "pest.h"
#include "spec.h"
#include "log.h"
#include "io.h"
#include "string.h"
#include "def.h"
#include "state.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "knobs.h"

#define KNOBINFO(N,S) irc_out("%-7s = %s",N,S);

static char const *knobtype_name[] = {
  [KNOB_BOOL]   = "bool",
  [KNOB_INT]    = "int",
  [KNOB_STR]    = "str",
};

static knob_t const *knob_lookup(char const *name){
  for(size_t x = 0;x < ARR_CNT(knob);x++){
    knob_t const *k = knob+x;
    if(strcmpci(k->name,name) == 0) return k;
  }
  irc_err("knob '%s' does not exist",name);
  return 0;
}

static void strval_bool(char *buf,size_t buf_size,bool const *data){
  snprintf(buf,buf_size,"%d",*data);
}

static void strval_int(char *buf,size_t buf_size,int64_t const *data){
  snprintf(buf,buf_size,"%ld",*data);
}

static void strval_str(char *buf,size_t buf_size,char const *const *data){
  strzcpy_quote(buf,*data,buf_size);
}

static unsigned setval_bool(knob_t const *k,char const *val){
  bool *data = k->data;

  if(!val){
    *data = k->def.b;
    return KNOB_SUCCESS;
  }

  if(val[0] == 0 || val[1] != 0){
    return KNOB_INVALID;
  }

  switch(val[0]){
  case '0': *data = 0;  return KNOB_SUCCESS;
  case '1': *data = 1;  return KNOB_SUCCESS;
  default:              return KNOB_INVALID;
  }
}

static bool char_num(char c){
  return c >= '0' && c <= '9';
}

static bool char_sign(char c){
  return c == '+' || c == '-';
}

static unsigned setval_int(knob_t const *k,char const *val){
  int64_t  *data = k->data;
  int64_t   v;

  if(!val){
    *data = k->def.i;
    return KNOB_SUCCESS;
  }

  if(!char_num(val[0]) && !char_sign(val[0])){
    return KNOB_INVALID;
  }

  for(char const *c = val+1;*c;c++){
    if(!char_num(*c)) return KNOB_INVALID;
  }

  v = strtoll(val,0,10);
  if(v < k->min || v > k->max){
    return KNOB_RANGE;
  }

  *data = v;
  return KNOB_SUCCESS;
}

static unsigned setval_str(knob_t const *k,char const *val){
  char    **data = k->data;
  int64_t   len;

  if(!val){
    strcpy(*data,k->def.s);
    return KNOB_SUCCESS;
  }

  len = strlen(val); 
  if(len < k->min || len > k->max){
    return KNOB_RANGE;
  }

  strcpy(*data,val);
  return KNOB_SUCCESS;
}

static bool storeval_bool(knob_t const *k,char const *path){
  bool const *v = k->data;
  return bool_store(path,*v) ? KNOB_SUCCESS : KNOB_STORE;
}

static unsigned loadval_bool(knob_t const *k,char const *path){
  return bool_load(path,k->data) ? KNOB_SUCCESS : KNOB_LOAD;
}

static bool storeval_int(knob_t const *k,char const *path){
  int64_t const *v = k->data;
  return int_store(path,*v) ? KNOB_SUCCESS : KNOB_STORE;
}

static unsigned loadval_int(knob_t const *k,char const *path){
  char buf[INT64_LEN+1+1]; //str\0 +1 for detecting oversize

  if(!str_load(path,buf,sizeof(buf))){
    return KNOB_LOAD;
  }

  return setval_int(k,buf);
}

static bool storeval_str(knob_t const *k,char const *path){
  char const **v = k->data;
  return str_store(path,*v) ? KNOB_SUCCESS : KNOB_STORE;
}

static unsigned loadval_str(knob_t const *k,char const *path){
  char buf[k->max+1+1]; //str\0 +1 for detecting oversize

  if(!str_load(path,buf,sizeof(buf))){
    return KNOB_LOAD;
  }

  return setval_str(k,buf);
}

static strval_f knob_strval[] = {
  [KNOB_BOOL]   = (strval_f)    strval_bool,
  [KNOB_INT]    = (strval_f)    strval_int,
  [KNOB_STR]    = (strval_f)    strval_str,
};

static setval_f knob_setval[] = {
  [KNOB_BOOL]   = (setval_f)    setval_bool,
  [KNOB_INT]    = (setval_f)    setval_int,
  [KNOB_STR]    = (setval_f)    setval_str,
};

static storeval_f knob_storeval[] = {
  [KNOB_BOOL]   = (storeval_f)  storeval_bool,
  [KNOB_INT]    = (storeval_f)  storeval_int,
  [KNOB_STR]    = (storeval_f)  storeval_str,
};

static loadval_f knob_loadval[] = {
  [KNOB_BOOL]   = (loadval_f)   loadval_bool,
  [KNOB_INT]    = (loadval_f)   loadval_int,
  [KNOB_STR]    = (loadval_f)   loadval_str,
};

static void knob_print_val(knob_t const *k){
  char val[KNOB_VAL_BUF];
  knob_strval[k->type](val,sizeof(val),k->data);
  irc_out("%s = %s",k->name,val);
}

static void knob_print(knob_t const *k,bool alt){
  char val[KNOB_VAL_BUF];
  char def[KNOB_VAL_BUF];
  char min[KNOB_VAL_BUF];
  char max[KNOB_VAL_BUF];

  knob_strval[k->type](val,sizeof(val),k->data);
  knob_strval[k->type](def,sizeof(def),&k->def);

  switch(k->type){
  case KNOB_INT:
  case KNOB_STR:
    strval_int(min,sizeof(min),&k->min);
    strval_int(max,sizeof(max),&k->max);
    break;
  }

  if(!alt){
    irc_out("%s = %s",k->name,val);
    return;
  }

  irc_out("--- BEGIN KNOB INFO ---");

  KNOBINFO("name",        k->name);
  KNOBINFO("type",        knobtype_name[k->type]);
  KNOBINFO("descrip",     k->desc);
  KNOBINFO("value",       val);
  KNOBINFO("default",     def);

  switch(k->type){
  case KNOB_INT:
    if(k->min != KNOB_MIN) KNOBINFO("minimum",min);
    if(k->max != KNOB_MAX) KNOBINFO("maximum",max);
    break;

  case KNOB_STR:
    if(k->min != KNOB_MIN) KNOBINFO("min len",min);
    if(k->max != KNOB_MAX) KNOBINFO("max len",max);
    break;
  }

  irc_out("--- END KNOB INFO ---");
}

static void knob_desc(knob_t const *k){
  char val[KNOB_VAL_BUF];
  knob_strval[k->type](val,sizeof(val),k->data);
  irc_out("%-7s = %-20s %-5s %s",
          k->name,val,knobtype_name[k->type],k->desc);
}

void knob_display(char const *name,bool alt){
  if(name){
    knob_t const *k = knob_lookup(name);
    if(!k) return;
    knob_print(k,alt);
  }else{
    irc_out("--- BEGIN KNOB LIST ---");
    for(size_t x = 0;x < ARR_CNT(knob);x++){
      knob_t const *k = knob+x;
      knob_desc(k);
    }
    irc_out("--- END KNOB LIST ---");
  }
}

static void knob_err(knob_t const *k,unsigned rc,char const *val){
  switch(rc){
    bool min;
    bool max;

  case KNOB_INVALID:
    irc_err("invalid value '%s' for knob '%s'",
            val,k->name);
    break;

  case KNOB_RANGE:
    min = k->min != KNOB_MIN;
    max = k->max != KNOB_MAX;
    if(!min && !max){
      assert(0);
    }else if(!min && max){
      irc_err("knob '%s' expects value <= %ld",
              k->name,k->min,k->max);
    }else if(min && !max){
      irc_err("knob '%s' expects value >= %ld",
              k->name,k->min,k->max);
    }else if(min && max){
      irc_err("knob '%s' expects value in range [%ld,%ld]",
              k->name,k->min,k->max);
    }
    break;

  case KNOB_STORE:
    irc_err("couldn't store knob '%s' on disk",
            k->name);
    break;

  case KNOB_LOAD:
    irc_err("couldn't load knob '%s' from disk",
            k->name);
    break;
  }
}

static bool knob_storef(knob_t const *k){
  char      path[PATH_BUF];
  unsigned  rc;

  FMT_PATH(path,"%s",k->name);
  rc = knob_storeval[k->type](k,path);

  switch(rc){
  case KNOB_SUCCESS:
    return 1;

  default:
    knob_err(k,rc,0);
    return 0;
  }
}

static bool knob_loadf(knob_t const *k){
  char      path[PATH_BUF];
  unsigned  rc;

  FMT_PATH(path,"%s",k->name);
  if(!file_exists(path)){
    unsigned rc = knob_setval[k->type](k,0);
    assert(rc == KNOB_SUCCESS); (void)rc;
    return knob_storef(k);
  }

  rc = knob_loadval[k->type](k,path);

  switch(rc){
  case KNOB_SUCCESS:
    return 1;

  default:
    knob_err(k,rc,0);
    return 0;
  }
}

void knob_set(char const *name,char const *val){
  knob_t const *k;
  unsigned      rc;

  k = knob_lookup(name);
  if(!k) return;

  if(val[0] == '\\' && val[1] == 0) val = ""; //hak
  rc = knob_setval[k->type](k,val);

  switch(rc){
  case KNOB_SUCCESS:
    knob_storef(k);
    knob_print_val(k);
    break;

  default:
    knob_err(k,rc,val);
    break;
  }
}

bool knob_load(void){
  for(size_t x = 0;x < ARR_CNT(knob);x++){
    knob_t const *k = knob+x;
    if(!knob_loadf(k)) return 0;
  }
  return 1;
}

