#include "chain.h"
#include "io.h"
#include "log.h"
#include "util.h"

#define MSG_POW2                        MESS_MSG_POW2
#define MSG_MAX                         (1<<MSG_POW2)
#define MSG_FILE_SIZE                   (MSG_MAX*MSG_SIZE)
#define MSG_FILE                        "msg"
#define MSG_FILE_ZEROS                  MESS_MSG_ZEROS

//--- mess message -----------------------------------------------------------
#define MSG_MESSAGE_SIZE                MESSAGE_SIZE

#define MSG_MESSAGE_OFF                 ( 0                                   )

#define MSG_SIZE                        ( MSG_MESSAGE_OFF + MSG_MESSAGE_SIZE  )

static void msg_path(char *buf,size_t buf_size,mess_t const *c,size_t i){
  snprintf(buf,buf_size,"%s/"MSG_FILE"%0"STR(MSG_FILE_ZEROS)"zu",c->path,i);
}

static void msg_corrupt(char const *path,bool write){
  irclog_err("mess message file corrupt: '%s'",path);
  if(write) disk_unsync();
}

static FILE *msg_open(char const *path,char const *mode,bool write){
  FILE *f = file_open(path,mode);
  if(!f){
    if(write) disk_unsync();
  }
  return f;
}

static bool msg_close(FILE *f,char const *path,bool write){
  if(!file_close(f)){
    if(write) disk_unsync();
    return 0;
  }
  if(write){
    disk_written(path);
  }else{
    disk_read(path);
  }
  return 1;
}

static bool msg_seek_entry(FILE *f,size_t index,bool write){
  if(!file_seek(f,index*MSG_SIZE,SEEK_SET)){
    file_close(f);
    if(write) disk_unsync();
    return 0;
  }
  return 1;
}

static bool msg_read_message(FILE *f,uint8_t *m,bool write){
  if(m){
    if(!file_read(f,m,MSG_MESSAGE_SIZE)){
      file_close(f);
      if(write) disk_unsync();
      return 0;
    }
  }else{
    if(!file_seek(f,MSG_MESSAGE_SIZE,SEEK_CUR)){
      file_close(f);
      if(write) disk_unsync();
      return 0;
    }
  }
  return 1;
}

static bool msg_write_message(FILE *f,uint8_t const *m){
  if(!file_write(f,m,MSG_MESSAGE_SIZE)){
    file_close(f);
    disk_unsync();
    return 0;
  }
  return 1;
}

bool msg_cnt(mess_t *c,size_t *msg_cnt_out){
  size_t msg_cnt = 0;
  while(1){
    char path[PATH_BUF];
    msg_path(path,sizeof(path),c,msg_cnt);
    if(!file_exists(path)) break;
    msg_cnt++;
  }
  *msg_cnt_out = msg_cnt;
  return 1;
}

bool msg_ecnt(mess_t *c,size_t *ecnt_out){
  char        path[PATH_BUF];
  struct stat sb;
  size_t      size;

  msg_path(path,sizeof(path),c,c->msg_cnt-1);
  if(c->msg_cnt == 0 || !file_exists(path)){
    *ecnt_out = 0;
    return 1;
  }
  if(!file_stat(path,&sb)) return 0;

  size = (size_t)sb.st_size;
  if(size % MSG_SIZE != 0 ||
     size > MSG_SIZE*MSG_MAX){
    msg_corrupt(path,0);
    return 0;
  }

  *ecnt_out = size/MSG_SIZE;
  return 1;
}

static inline size_t msg_file_idx(size_t index){
  return index>>MSG_POW2;
}

static inline size_t msg_file_off(size_t index){
  return index & MSG_MAX-1;
}

static inline size_t msg_index_next(mess_t const *c){
  return (c->msg_cnt-1)*MSG_MAX+c->msg_ecnt;
}

size_t msg_count(mess_t const *c){
  return c->msg_cnt > 0 ? msg_index_next(c) : 0;
}

bool msg_ins(mess_t        *c,
             uint8_t const *m,
             uint64_t      *index){

  char  path[PATH_BUF];
  FILE *f;

  if(c->msg_cnt == 0 || c->msg_ecnt+1 > MSG_MAX){
    c->msg_cnt++;
    c->msg_ecnt = 0;
  }

  msg_path(path,sizeof(path),c,c->msg_cnt-1);
  f = msg_open(path,"ab",1);
  if(!f) return 0;

  if(!msg_write_message (f,m))          return 0;
  if(!msg_close         (f,path,1))     return 0;

  *index = msg_index_next(c);
  c->msg_ecnt++;

  return 1;
}

bool msg_get(mess_t    *c,
             uint8_t   *m,
             uint64_t   index){

  char    path[PATH_BUF];
  size_t  file_idx;
  size_t  file_off;
  FILE   *f;

  if(!m) return 1;

  file_idx = msg_file_idx(index);
  file_off = msg_file_off(index);

  msg_path(path,sizeof(path),c,file_idx);
  f = msg_open(path,"rb",0);
  if(!f) return 0;

  //TODO: chek filesize

  if(!msg_seek_entry    (f,file_off,0)) return 0;
  if(!msg_read_message  (f,m,0))        return 0;
  if(!msg_close         (f,path,0))     return 0;

  return 1;
}

