#include "net.h"
#include "def.h"
#include "log.h"
#include "crypto.h"
#include "util.h"
#include "time.h"
#include "config.h"
#include "state.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>

bool ip_u32(uint32_t *ip,char const *str){
  struct sockaddr_in sa;
  if(inet_pton(AF_INET,str,&sa.sin_addr) != 1){
    log_errno("inet_pton");
    return 0;
  }
  *ip = ntohl(sa.sin_addr.s_addr);
  return 1;
}

bool ip_str(char *str,size_t str_size,uint32_t ip){
  struct sockaddr_in sa;
  VAR_ZERO(sa);
  sa.sin_addr.s_addr = htonl(ip);
  if(inet_ntop(AF_INET,&sa.sin_addr,str,str_size) == 0){
    log_errno("inet_ntop");
    return 0;
  }
  return 1;
}

bool udp_bound(void){
  return udp_fd != -1;
}

size_t udp_recv(uint32_t *src_ip,
                uint16_t *src_port,
                uint16_t  port,
                uint8_t  *data,
                size_t    size){

  struct sockaddr_in  addr;
  struct sockaddr_in  cli_addr;
  socklen_t           cli_addr_len;
  ssize_t             n;

  if(udp_fd == -1){
    udp_fd = socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP);
    if(udp_fd == -1){
      log_errno("socket");
      return 0;
    }

    if(fcntl(udp_fd,F_SETFL,O_NONBLOCK) != 0){
      log_errno("fcntl");
      close(udp_fd);
      udp_fd = -1;
      return 0;
    }

    VAR_ZERO(addr);
    addr.sin_family       = AF_INET;
    addr.sin_port         = htons(port);
    addr.sin_addr.s_addr  = htonl(INADDR_ANY);

    if(bind(udp_fd,
            (struct sockaddr *)&addr,
            sizeof(addr)) != 0){
      log_errno("bind");
      close(udp_fd);
      udp_fd = -1;
      sleep_nsec(NET_RETRY_SLEEP*NSEC_MSEC);
      return 0;
    }
    log_out("listening on UDP port %u",port);
  }

  if(data == 0) return 0;

  cli_addr_len  = sizeof(cli_addr);
  n             = recvfrom(udp_fd,
                           data,
                           size,
                           0,
                           (struct sockaddr *)&cli_addr,
                           &cli_addr_len);

  if(n == -1){
    if(errno == EAGAIN || errno == EWOULDBLOCK){
      udp_silent = 1;
      return 0;
    }
    log_errno("recvfrom");
    return 0;
  }
  udp_silent = 0;

  *src_ip   = ntohl(cli_addr.sin_addr.s_addr);
  *src_port = ntohs(cli_addr.sin_port);
  return n;
}

size_t udp_send(uint32_t        ip,
                uint16_t        port,
                uint8_t const  *data,
                size_t          size){

  struct sockaddr_in  addr;
  ssize_t             n;

  if(udp_fd == -1) return 0;

  VAR_ZERO(addr);
  addr.sin_family       = AF_INET;
  addr.sin_port         = htons(port);
  addr.sin_addr.s_addr  = htonl(ip);

  n = sendto(udp_fd,  
             data,
             size,
             0,
             (struct sockaddr const *)&addr,
             sizeof(addr));

  if(n == -1){
    log_errno("sendto");
    return 0;
  }
  return n;
}

void tcp_init(tcpev_f func){
  tcp_func = func;
}

size_t tcp_recv(uint32_t *src_ip,
                uint16_t *src_port,
                uint16_t  port,
                uint8_t  *data,
                size_t    size){

  struct sockaddr_in  addr;
  struct sockaddr_in  cli_addr;
  socklen_t           cli_addr_len;
  ssize_t             n;

  if(tcp_listen_fd == -1){
    tcp_listen_fd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if(tcp_listen_fd == -1){
      log_errno("socket");
      return 0;
    }

    int enable = 1;
    if(setsockopt(tcp_listen_fd,
                  SOL_SOCKET,
                  SO_REUSEADDR,
                  &enable,
                  sizeof(int)) == -1){
      log_errno("setsockopt");
    }

    if(fcntl(tcp_listen_fd,F_SETFL,O_NONBLOCK) != 0){
      log_errno("fcntl");
      tcp_lclose();
      sleep_nsec(NET_RETRY_SLEEP*NSEC_MSEC);
      return 0;
    }

    VAR_ZERO(addr);
    addr.sin_family       = AF_INET;
    addr.sin_port         = htons(port);
    addr.sin_addr.s_addr  = htonl(INADDR_ANY);

    if(bind(tcp_listen_fd,
            (struct sockaddr *)&addr,
            sizeof(addr)) != 0){
      log_errno("bind");
      tcp_lclose();
      sleep_nsec(NET_RETRY_SLEEP*NSEC_MSEC);
      return 0;
    }

    if(listen(tcp_listen_fd,10) != 0){
      log_errno("listen");
      tcp_lclose();
      sleep_nsec(NET_RETRY_SLEEP*NSEC_MSEC);
      return 0;
    }
    log_out("listening on TCP port %u",port);
  }

  if(tcp_client_fd == -1){
    char ips[IPV4_BUF];

    cli_addr_len  = sizeof(cli_addr);
    tcp_client_fd = accept(tcp_listen_fd,
                           (struct sockaddr *)&cli_addr,
                           &cli_addr_len);
    if(tcp_client_fd == -1){
      if(errno == EAGAIN || errno == EWOULDBLOCK){
        if(udp_silent){
          sleep_nsec(NET_SILENT_SLEEP*NSEC_MSEC);
        }
        return 0;
      }
      log_errno("accept");
      return 0;
    }

    if(fcntl(tcp_client_fd,F_SETFL,O_NONBLOCK) != 0){
      log_errno("fcntl");
      close(tcp_client_fd);
      tcp_client_fd = -1;
      return 0;
    }

    tcp_client_ip   = ntohl(cli_addr.sin_addr.s_addr);
    tcp_client_port = ntohs(cli_addr.sin_port);
    *src_ip         = tcp_client_ip;
    *src_port       = tcp_client_port;

    ip_str(ips,sizeof(ips),tcp_client_ip);
    log_out("accepted TCP client from %s:%u",
            ips,tcp_client_port);
    if(tcp_func) tcp_func(TCP_CONNECT);
  }else{
    *src_ip         = tcp_client_ip;
    *src_port       = tcp_client_port;
  }

  n = recv(tcp_client_fd,data,size,0);
  if(n == -1){
    if(errno == EAGAIN || errno == EWOULDBLOCK){
      tcp_silent = 1;
      if(udp_silent){
        sleep_nsec(NET_SILENT_SLEEP*NSEC_MSEC);
      }
      return 0;
    }
    log_errno("recv");
    tcp_cclose();
    return 0;
  }else if(n == 0){
    tcp_cclose();
    return 0;
  }

  tcp_silent = 0;
  return n;
}

size_t tcp_send(uint8_t const *data,size_t size){
  ssize_t n;

  if(tcp_client_fd == -1) return 0;

  n = send(tcp_client_fd,data,size,0);
  if(n == -1){
    log_errno("send");
    tcp_cclose();
    return 0;
  }

  return n;
}

void tcp_lclose(void){
  if(tcp_listen_fd == -1) return;
  close(tcp_listen_fd);
  tcp_listen_fd = -1;
  log_out("TCP listening socket closed");
}

void tcp_cclose(void){
  if(tcp_client_fd == -1) return;
  close(tcp_client_fd);
  tcp_client_fd = -1;
  log_out("TCP client socket closed");
  if(tcp_func) tcp_func(TCP_DISCONNECT);
}

