#include "log.h"
#include "term.h"
#include "addr.h"
#include "crypto.h"
#include "time.h"
#include "config.h"
#include <stdio.h>
#include <assert.h>

#define ARROW_DISCARD         ANSI_GRAY  "<-" ANSI_NORM
#define ARROW_IN              ANSI_GREEN "<-" ANSI_NORM
#define ARROW_OUT             ANSI_RED   "->" ANSI_NORM

static char const *cmd_abbrev[] = {
  [ CMD_BROADCAST   ] = "BC",  
  [ CMD_DIRECT      ] = "DM",  
  [ CMD_BANNER      ] = "BA",  
  [ CMD_GETDATA     ] = "GD",  
  [ CMD_KEYOFFER    ] = "KO",  
  [ CMD_KEYSLICE    ] = "KS",  
  [ CMD_ADDRESSCAST ] = "AC",  
  [ CMD_IGNORE      ] = "IG",  
};

void log_packet(hash_t const  hash,
                uint8_t       cmd,
                peer_t       *p,
                addr_t const *addr,
                bool          outgoing,
                uint64_t      time,
                uint64_t      stamp){

  char        hash_hex [HASH_HEX_BUF];
  char        at       [AT_BUF];
  char        tdiff    [STAMPDIFF_BUF];
  char const *dst = p ? p->handle[0] : "*";

#ifndef LOG_IGNORE
  if(cmd == CMD_IGNORE) return;
#endif

  hash_str_log(hash_hex,sizeof(hash_hex),hash);
  if(addr){
    at_from_addr(at,sizeof(at),addr);
  }else{
    *at = 0;
  }

  if(outgoing){
    log_out("%s %s " ARROW_OUT " %s %s",
            cmd_abbrev[cmd],hash_hex,dst,at);
  }else{
    fmt_stampdiff(tdiff,sizeof(tdiff),time,stamp);
    log_out("%s %s " ARROW_IN " %s %s [%s]",
            cmd_abbrev[cmd],hash_hex,dst,at,tdiff);
  }
}

void log_martian(addr_t const *addr,bool reason_size){
  char at[AT_BUF];

#ifndef LOG_MARTIAN
  return;
#endif

  at_from_addr(at,sizeof(at),addr);
  if(reason_size){
    log_out("?? [  ??  ] " ARROW_DISCARD " %s [martian: size != %d]",
            at,BLACK_SIZE);
  }else{
    log_out("?? [  ??  ] " ARROW_DISCARD " %s [martian: no key]",
            at);
  }
}

static void log_discard(hash_t const   hash,
                        uint8_t        cmd,
                        peer_t        *p,
                        addr_t const  *addr,
                        uint64_t       time,
                        uint64_t       stamp,
                        char const    *reason){

  char hash_hex [HASH_HEX_BUF];
  char at       [AT_BUF];
  char tdiff    [STAMPDIFF_BUF];

  hash_str_log(hash_hex,sizeof(hash_hex),hash);
  at_from_addr(at,sizeof(at),addr);
  fmt_stampdiff(tdiff,sizeof(tdiff),time,stamp);

  log_out("%s %s " ARROW_DISCARD " %s %s [%s] [%s]",
          cmd_abbrev[cmd],hash_hex,p->handle[0],at,reason,tdiff);
}

void log_stale(hash_t const   hash,
               uint8_t        cmd,
               peer_t        *p,
               addr_t const  *addr,
               uint64_t       time,
               uint64_t       stamp){

#ifndef LOG_STALE
  return;
#endif

  log_discard(hash,cmd,p,addr,time,stamp,"stale");
}

void log_dupe(hash_t const   hash,
              uint8_t        cmd,
              peer_t        *p,
              addr_t const  *addr,
              uint64_t       time,
              uint64_t       stamp){

#ifndef LOG_DUPLICATE
  return;
#endif

  log_discard(hash,cmd,p,addr,time,stamp,"duplicate");
}

void log_dedup_full(hash_t const   hash,
                    uint8_t        cmd,
                    peer_t        *p,
                    addr_t const  *addr,
                    uint64_t       time,
                    uint64_t       stamp){

#ifndef LOG_DUPLICATE
  return;
#endif

  log_discard(hash,cmd,p,addr,time,stamp,"dedup buffer full");
}

void log_cutoff(hash_t const  hash,
                uint8_t       cmd,
                peer_t       *p,
                addr_t const *addr,
                uint64_t      time,
                uint64_t      stamp,
                uint8_t       bounces){

#ifndef LOG_CUTOFF
  return;
#endif

  char reason[21+INT64_LEN+1]; //discarded: bounces = $num\0
  snprintf(reason,sizeof(reason),"discarded: bounces = %u",bounces);
  log_discard(hash,cmd,p,addr,time,stamp,reason);
}

