#include "peer.h"
#include "addr.h"
#include "chain.h"
#include "crypto.h"
#include "pest.h"
#include "log.h"
#include "io.h"
#include "time.h"
#include "util.h"
#include "string.h"
#include "wipe.h"
#include "state.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

peer_t *peer_lookup_id(peerid_t id){
  if(id == PEER_ID_INVALID) return 0;
  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p = peer+x;
    if(p->id == id) return p;
  }
  return 0;
}

peer_t *peer_lookup_noerr(char const *handle){
  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p = peer+x;
    for(size_t y = 0;y < p->handle_cnt;y++){
      if(strncmp(handle,p->handle[y],HANDLE_LEN) == 0) return p;
    }
  }
  return 0;
}

peer_t *peer_lookup(char const *handle){
  peer_t *p = peer_lookup_noerr(handle);
  if(!p){
    irc_err("peer '%s' not found in WOT",handle);
  }
  return p;
}

static void peer_path(peer_t *p){
  char  path[PATH_BUF];
  char *handle = p->handle[0];

  FMT_PATH(path,DIR_PEER"/%s/"DIR_MESS,handle);
  mess_path(&p->mess,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_DCHAIN,handle);
  chain_path(&p->dchain,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_BCHAIN,handle);
  chain_path(&p->bchain,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_DCHAIN"/"FILE_HEAD,handle);
  ht_path(&p->dhead,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_DCHAIN"/"FILE_TAIL,handle);
  ht_path(&p->dtail,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_BCHAIN"/"FILE_HEAD,handle);
  ht_path(&p->bhead,path);

  FMT_PATH(path,DIR_PEER"/%s/"DIR_BCHAIN"/"FILE_TAIL,handle);
  ht_path(&p->btail,path);
}

static void peer_init(peer_t *p,char const *handle,peerid_t id){
  strzcpy(p->handle[0],handle,sizeof(p->handle[0]));

  p->id         = id;
  p->handle_cnt = 1;
  p->key_cnt    = 0;
  p->addr.ip    = 0;
  p->addr.port  = 0;
  p->stamp      = 0;
  p->pak_stamp  = 0;
  p->con_stamp  = mono_sec();
  p->active     = 0;
  p->paused     = 0;
  p->banner[0]  = 0;

  mess_init(&p->mess);
  chain_init(&p->dchain);
  chain_init(&p->bchain);
  ht_init(&p->dhead);
  ht_init(&p->dtail);
  ht_init(&p->bhead);
  ht_init(&p->btail);
  hash_zero(p->self);
  rekey_init(&p->rekey);

  peer_path(p);
}

static uint16_t id_next(void){
  peerid_t end = peer_id-1;
  while(peer_lookup_id(peer_id)){
    if(peer_id == end)              return PEER_ID_INVALID;
    if(peer_id == PEER_ID_INVALID)  return PEER_ID_INVALID;
  }
  return peer_id++;
}

peer_t *peer_add_internal(char const *handle,bool boot){
  log_f     out = boot ? log_void : irclog_out;
  log_f     err = boot ? log_err  : irc_err;
  peer_t   *p;
  peerid_t  id;

  if(peer_cnt >= PEER_MAX){
    err("max peers reached, increase dunbar number in source");
    return 0;
  }

  //TODO: more useful eror, maxlen, chars alowed osv
  if(!handle_valid(handle)){
    err("invalid peer handle: '%s'",handle);
    return 0;
  }

  if(strncmp(handle,self_handle,HANDLE_LEN) == 0){
    err("peer handle cannot be same as own handle");
    return 0;
  }

  if(peer_lookup_noerr(handle)){
    err("handle '%s' already exists in WOT",handle);
    return 0;
  }

  id = id_next();
  if(id == PEER_ID_INVALID){
    err("peer ids exhausted");
    return 0;
  }

  p = &peer[peer_cnt++];
  peer_init(p,handle,id);
  if(!peer_loadpeer(p)) return 0;

  out("peer '%s' added",handle);
  return p;
}

peer_t *peer_add(char const *handle){
  return peer_add_internal(handle,0);
}

void peer_del(peer_t *p){
  char  path[PATH_BUF];
  char  handle[HANDLE_BUF];

  hearsay_del_peer(p);
  pending_del_peer(p);
  getdata_del_peer(p);

  strzcpy(handle,p->handle[0],sizeof(handle));
  ARR_DEL(peer,peer_cnt,p);
  peer_cnt--;
  FMT_PATH(path,DIR_PEER"/%s",handle);
  dir_zero_nolog(path);
  if(!rm_rf(path)){
    disk_unsync();
  }
  irclog_out("peer '%s' deleted",handle);
}

bool peer_alias_add_internal(peer_t *p,char const *alias,bool boot){
  log_f   out = boot ? log_void : irclog_out;
  log_f   err = boot ? log_err  : irc_err;
  char   *handle = p->handle[0];

  if(p->handle_cnt >= HANDLE_MAX){
    err("max number of aliases (%d) reached for peer '%s'",HANDLE_MAX,handle);
    return 0;
  }

  if(!handle_valid(alias)){
    err("invalid peer alias: '%s'",alias);
    return 0;
  }

  if(strncmp(alias,self_handle,HANDLE_LEN) == 0){
    err("peer alias cannot be same as own handle");
    return 0;
  }

  if(peer_lookup_noerr(alias)){
    err("handle '%s' already exists in WOT",alias);
    return 0;
  }

  strzcpy(p->handle[p->handle_cnt],alias,sizeof(p->handle[p->handle_cnt]));
  p->handle_cnt++;

  if(!boot) peer_alias_store(p);

  out("alias '%s' added for peer '%s'",alias,handle);
  return 1;
}

bool peer_alias_add(peer_t *p,char const *alias){
  return peer_alias_add_internal(p,alias,0);
}

unsigned peer_alias_del(peer_t *p,char const *alias){
  char    handle_old[HANDLE_BUF];
  char   *h;
  bool    rename;

  strzcpy(handle_old,p->handle[0],sizeof(handle_old));

  if(p->handle_cnt == 1){
    irc_err("cannot delete last handle for peer '%s'",handle_old);
    return 0;
  }

  h = peer_handle_lookup(p,alias);
  assert(h);

  rename = h == p->handle[0];
  ARR_DEL(p->handle,p->handle_cnt,(handle_t *)h);
  p->handle_cnt--;

  irclog_out("alias '%s' for peer '%s' deleted",alias,handle_old);

  if(rename){
    char  path_from  [PATH_BUF];
    char  path_to    [PATH_BUF];
    char *handle_new;

    handle_new = p->handle[0];

    FMT_PATH(path_from, DIR_PEER"/%s",handle_old);
    FMT_PATH(path_to,   DIR_PEER"/%s",handle_new);
    if(!file_rename(path_from,path_to)){
      disk_unsync();
    }
    peer_path(p);

    irclog_out("main handle for peer '%s' is now '%s'",handle_old,handle_new);
  }

  peer_alias_store(p);
  return rename ? 2 : 1;
}

static void pkey_init(pkey_t *pk,uint8_t const *key,uint64_t time){
  memcpy(pk->key.full,key,KEY_SIZE);
  serpent_init(pk->ciph_serp,pk->key.sub.ciph,sizeof(pk->key.sub.ciph));
  pk->time = time;
}

pkey_t *peer_lookup_pkey(peer_t *p,uint8_t const *key){
  for(size_t x = 0;x < p->key_cnt;x++){
    pkey_t *pk = p->key+x;
    if(memcmp(pk->key.full,key,KEY_SIZE) == 0){
      return pk;
    }
  }
  return 0;
}

pkey_t *pkey_lookup_noerr(uint8_t const *key,peer_t **peer_out){
  if(peer_out) *peer_out = 0;
  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p   = peer+x;
    pkey_t *pk  = peer_lookup_pkey(p,key);
    if(pk){
      if(peer_out) *peer_out = p;
      return pk;
    }
  }
  return 0;
}

static pkey_t *pkey_lookup(uint8_t const *key,peer_t **peer_out){
  pkey_t *pk = pkey_lookup_noerr(key,peer_out);
  if(!pk){
    char b64_key[B64_KEY_BUF];
    b64_from_key(b64_key,sizeof(b64_key),key);
    irc_err("key '%s' not found in WOT",b64_key);
    BUF_ZERO(b64_key);
  }
  return pk;
}

bool peer_key_add_internal(peer_t        *p,
                           uint8_t const *key,
                           uint64_t       time,
                           bool           boot,
                           bool           silent){

  log_f   out = boot || silent ? log_void : irclog_out;
  log_f   err = boot           ? log_err  : irc_err;
  char   *handle = p->handle[0];
  char    b64_key[B64_KEY_BUF];

  if(p->key_cnt >= KEY_MAX){
    err("max number of keys (%d) reached for peer '%s'",KEY_MAX,handle);
    return 0;
  }

  b64_from_key(b64_key,sizeof(b64_key),key);

  if(entropy_missing(key,sizeof(key))){
    err("entropy obviously missing from key: '%s'",b64_key);
    BUF_ZERO(b64_key);
    return 0;
  }

  if(pkey_lookup_noerr(key,0)){
    err("key '%s' already exists in WOT",b64_key);
    BUF_ZERO(b64_key);
    return 0;
  }

  pkey_init(&p->key[p->key_cnt],key,time);
  p->key_cnt++;

  if(!boot) peer_key_store(p);

  out("key '%s' added for peer '%s'",b64_key,handle);
  BUF_ZERO(b64_key);
  return 1;
}

bool peer_key_add(peer_t *p,uint8_t const *key,bool silent){
  return peer_key_add_internal(p,key,0,0,silent);
}

bool peer_key_del(uint8_t const *key,bool silent){
  log_f     out = silent ? log_void : irclog_out;
  peer_t   *p;
  pkey_t   *pk;
  char      b64_key[B64_KEY_BUF];
  char     *handle;

  pk = pkey_lookup(key,&p);
  if(!pk) return 0;
  handle = p->handle[0];

  if(p->key_cnt <= 1){
    irc_err("cannot delete last key for peer '%s'",handle);
    return 0;
  }

  ARR_DEL(p->key,p->key_cnt,pk);
  p->key_cnt--;
  memset_ffs(p->key+p->key_cnt,0x00,sizeof(pkey_t));

  peer_key_store(p);

  b64_from_key(b64_key,sizeof(b64_key),key);
  out("key '%s' for peer '%s' deleted",b64_key,handle);
  BUF_ZERO(b64_key);
  return 1;
}

void peer_key_stamp(peer_t *p,pkey_t *pk,uint64_t time){
  pk->time = time;
  peer_key_store(p);
}

void peer_banner_set(peer_t *p,char const *banner){
  strzcpy(p->banner,banner,sizeof(p->banner));
  peer_banner_store(p);
}

pkey_t *peer_key_recent(peer_t *p){
  pkey_t   *k   = 0;
  uint64_t  max = 0;
  for(size_t x = 0;x < p->key_cnt;x++){
    pkey_t *pk = p->key+x;
    //prefer higher index so uses gorekt key
    //when rekey even if stamp identical
    if(pk->time >= max){
      k   = pk;
      max = pk->time;
    }
  }

  if(!k){
    irc_err("no key exists for peer '%s'",p->handle[0]);
  }

  return k;
}

bool peer_addr_set(peer_t *p,addr_t *addr,bool spam){
  log_f out = spam ? irclog_out : log_out;
  char  at[AT_BUF];

  assert(addr_valid(addr));
  p->addr = *addr;

  peer_addr_store(p);

  at_from_addr(at,sizeof(at),addr);
  out("address for peer '%s' set to '%s'",p->handle[0],at);
  return 1;
}

void peer_pause(peer_t *p,bool pause){
  p->paused = pause;
  peer_pause_store(p);
  irclog_out("peer '%s' %spaused",p->handle[0],p->paused ? "" : "un");
}

void peer_stamp(peer_t *p,uint64_t stamp){
  p->stamp = stamp;
  peer_stamp_store(p);
}

void peer_pak_stamp(peer_t *p,uint64_t stamp){
  p->pak_stamp = stamp;
}

char *peer_handle_lookup(peer_t *p,char const *handle){
  for(size_t x = 0;x < p->handle_cnt;x++){
    char *h = p->handle[x];
    if(strncmp(handle,h,HANDLE_LEN) == 0) return h;
  }
  return 0;
}

bool peer_cansend(peer_t *p,bool error){
  log_f   err     = error ? irclog_err : log_void;
  char   *handle  = p->handle[0];

  if(p->key_cnt == 0){
    err("cannot send to peer '%s', no key exists ",handle);
    return 0;
  }

  if(p->addr.ip == 0){
    err("cannot send to peer '%s', no ip exists",handle);
    return 0;
  }

  if(p->addr.port == 0){
    err("cannot send to peer '%s', no port exists",handle);
    return 0;
  }

  if(p->paused){
    err("cannot send to peer '%s', peer is paused",handle);
    return 0;
  }

  return 1;
}

//TODO: fix this great stuf
static peer_t *pkey_index_cmp_peer = 0;
static int pkey_index_cmp(size_t const *a,size_t const *b){
  peer_t *p = pkey_index_cmp_peer;
  return  (p->key[*a].time < p->key[*b].time)
         -(p->key[*a].time > p->key[*b].time);
}

static void peer_print_internal(peer_t *p,bool key){
  char        handles [(HANDLE_LEN+2)*HANDLE_MAX]; //handle (alias0, alias1)\0
  char        at      [AT_BUF];
  char        forks   [9];  //99 forks\0
  char        breaks  [10]; //99 breaks\0
  char        ts      [STAMP_BUF];
  char        b64_key [B64_KEY_BUF];
  size_t      f;
  size_t      b;

  char       *a = handles;
  char       *e = handles+sizeof(handles);
  char const *ps;

  a = strpcpy(a,p->handle[0],e);
  if(p->handle_cnt > 1){
    a = strpcpy(a," (",e);
    for(size_t x = 1;x < p->handle_cnt;x++){
      a = strpcpy(a,p->handle[x],e);
      if(x+1 < p->handle_cnt){
        a = strpcpy(a,", ",e);
      }
    }
    a = strpcpy(a,")",e);
  }

  if(!addr_null(&p->addr)){
    at_from_addr(at,sizeof(at),&p->addr);
  }else{
    *at = 0;
  }

  f   = 0;
  b   = 0;
  f  += ht_forks  (&p->bhead,&p->btail);
  b  += ht_breaks (&p->bhead,&p->btail);
  f  += ht_forks  (&p->dhead,&p->dtail);
  b  += ht_breaks (&p->dhead,&p->dtail);
  snprintf(forks, sizeof(forks),  "%2zu fork%c",  f,f == 1 ? ' ' : 's');
  snprintf(breaks,sizeof(breaks), "%2zu break%c", b,b == 1 ? ' ' : 's');

  if(p->stamp){
    time_str(ts,sizeof(ts),CONSOLE_STAMP_FMT,p->stamp);
  }else{
    *ts = 0;
  }

  ps = p->paused ? "paused" : "";

  if(key) irc_out("--- BEGIN PEER INFO ---");
  irc_out("%-24s  %-21s  %zu key%c  %-8s  %-9s  %-6s     %s",
          handles,at,p->key_cnt,p->key_cnt == 1 ? ' ' : 's',forks,breaks,ps,ts);
  if(!key) return;

  size_t index[KEY_MAX];
  for(size_t x = 0;x < p->key_cnt;x++){
    index[x] = x;
  }
  pkey_index_cmp_peer = p;
  qsort(index,p->key_cnt,sizeof(*index),(qsort_f)pkey_index_cmp);

  for(size_t x = 0;x < p->key_cnt;x++){
    pkey_t *pk = p->key+index[x];
    b64_from_key(b64_key,sizeof(b64_key),pk->key.full);
    if(pk->time){
      *ts = ' ';
      time_str(ts+1,sizeof(ts)-1,CONSOLE_STAMP_FMT,pk->time);
    }else{
      *ts = 0;
    }
    irc_out("%s%s",b64_key,ts);
  }
  BUF_ZERO(b64_key);

  if(*p->banner){
    irc_out("banner: '%s'",p->banner);
  }

  irc_out("--- END PEER INFO ---");
}

void peer_print(peer_t *p){
  peer_print_internal(p,1);
}

void peer_wot_print(void){
  if(peer_cnt == 0){
    irc_out("web of trust is empty");
    return;
  }
  irc_out("--- BEGIN WEB OF TRUST ---");
  for(size_t x = 0;x < peer_cnt;x++){
    peer_print_internal(peer+x,0);
  }
  irc_out("--- END WEB OF TRUST ---");
}

void peer_addr_print(peer_t *p){
  //TODO: "at" -> something else
  char at[AT_BUF];

  if(addr_null(&p->addr)) return;
  at_from_addr(at,sizeof(at),&p->addr);
  irc_out("%-24s  %s",p->handle[0],at);
}

static bool at_empty(void){
  for(size_t x = 0;x < peer_cnt;x++){
    if(!addr_null(&peer[x].addr)) return 0;
  }
  return 1;
}

void peer_at_print(void){
  if(at_empty()){
    irc_out("address table is empty");
    return;
  }
  irc_out("--- BEGIN ADDRESS TABLE ---");
  for(size_t x = 0;x < peer_cnt;x++){
    peer_addr_print(peer+x);
  }
  irc_out("--- END ADDRESS TABLE ---");
}

