#include "peer.h"
#include "addr.h"
#include "chain.h"
#include "crypto.h"
#include "io.h"
#include "log.h"
#include "util.h"
#include "string.h"
#include "state.h"
#include <dirent.h>
#include <stdlib.h>
#include <string.h>

bool peer_alias_store(peer_t *p){
  char  path[PATH_BUF];
  FILE *f;

  FMT_PATH(path,DIR_PEER"/%s/"FILE_ALIAS,p->handle[0]);
  if(p->handle_cnt <= 1){
    if(!file_delete_zero(path)){
      disk_unsync();
      return 0;
    }
    return 1;
  }
  f = file_open_zero(path,"wb");
  if(!f){
    disk_unsync();
    return 0;
  }

  for(size_t x = 1;x < p->handle_cnt;x++){
    if(!file_printf(f,"%s\n",p->handle[x])){
      file_close(f);
      disk_unsync();
      return 0;
    }
  }

  if(!file_close(f)){
    disk_unsync();
    return 0;
  }
  disk_written(path);
  return 1;
}

static bool peer_alias_load(peer_t *p){
  char    path[PATH_BUF];
  FILE   *f;
  char    buf[HANDLE_BUF+1+1]; //\n +1 for detecting invalid handle
  char   *b;

  FMT_PATH(path,DIR_PEER"/%s/"FILE_ALIAS,p->handle[0]);
  if(!file_exists(path)){
    return file_touch(path);
  }
  f = file_open(path,"rb");
  if(!f) return 0;

  while(file_gets(f,buf,sizeof(buf))){
    b = str_trim(buf);
    if(!*b) continue;

    if(!peer_alias_add_internal(p,b,1)){
      file_close(f);
      return 0;
    }
  }

  if(ferror(f)){
    file_close(f);
    return 0;
  }

  file_close(f);
  disk_read(path);
  return 1;
}

bool peer_key_store(peer_t *p){
  char  path[PATH_BUF];
  FILE *f;

  assert(p->key_cnt > 0);

  FMT_PATH(path,DIR_PEER"/%s/"FILE_KEY,p->handle[0]);
  f = file_open_zero(path,"wb");
  if(!f){
    disk_unsync();
    return 0;
  }

  for(size_t x = 0;x < p->key_cnt;x++){
    pkey_t *pk = p->key+x;
    char    b64_key[B64_KEY_BUF];
    bool    r;

    b64_from_key(b64_key,sizeof(b64_key),pk->key.full);
    if(pk->time > 0) r = file_printf(f,"%s %zu\n",b64_key,pk->time);
    else             r = file_printf(f,"%s\n",b64_key);
    BUF_ZERO(b64_key);
    if(!r){
      file_close(f);
      disk_unsync();
      return 0;
    }
  }

  if(!file_close(f)){
    disk_unsync();
    return 0;
  }
  disk_written(path);
  return 1;
}

static bool peer_key_load(peer_t *p){
  char    path[PATH_BUF];
  FILE   *f;
  char    buf[B64_KEY_LEN+WHITE_LEN+INT64_LEN+WHITE_LEN+1+1]; //key_timestamp\n\0

  FMT_PATH(path,DIR_PEER"/%s/"FILE_KEY,p->handle[0]);
  if(!file_exists(path)){
    return file_touch(path);
  }
  f = file_open(path,"rb");
  if(!f) return 0;

  while(file_gets(f,buf,sizeof(buf))){
    size_t    argc;
    char     *argv[2];
    uint64_t  time = 0;
    uint8_t   key[KEY_SIZE];
    char     *b64_key;

    argc = str_args(buf,2,argv);
    if(argc == 0) continue;
    if(argc > 2){
      log_err("invalid key file format: '%s'",path);
      file_close(f);
      return 0;
    }

    b64_key = argv[0];
    if(!key_from_b64(key,b64_key)){
      log_err("invalid key: '%s'",b64_key);
      key_zero(key);
      return 0;
    }

    if(argc > 1){
      time = strtoll(argv[1],0,10);
    }

    if(!peer_key_add_internal(p,key,time,1,0)){
      memset_ffs(b64_key,0x00,strlen(b64_key));
      key_zero(key);
      file_close(f);
      return 0;
    }

    memset_ffs(b64_key,0x00,strlen(b64_key));
    key_zero(key);
  }

  if(ferror(f)){
    file_close(f);
    return 0;
  }

  file_close(f);
  disk_read(path);
  return 1;
}

bool peer_addr_store(peer_t *p){
  char path[PATH_BUF];
  FMT_PATH(path,DIR_PEER"/%s/"FILE_ADDR,p->handle[0]);
  if(!addr_store(path,&p->addr)){
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool peer_addr_load(peer_t *p){
  char path[PATH_BUF];
  FMT_PATH(path,DIR_PEER"/%s/"FILE_ADDR,p->handle[0]);
  if(!file_exists(path)){
    return addr_store(path,&p->addr);
  }else{
    return addr_load(path,&p->addr);
  }
}

bool peer_stamp_store(peer_t *p){
  char path[PATH_BUF];
  char stamp[STAMP_BUF];

  snprintf(stamp,sizeof(stamp),"%lu",p->stamp);
  FMT_PATH(path,DIR_PEER"/%s/"FILE_STAMP,p->handle[0]);
  if(!str_store(path,stamp)){
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool peer_stamp_load(peer_t *p){
  char    path[PATH_BUF];
  int64_t i;

  FMT_PATH(path,DIR_PEER"/%s/"FILE_STAMP,p->handle[0]);
  if(!file_exists(path)){
    return int_store(path,p->stamp);
  }
  if(!int_load(path,&i)){
    return 0;
  }
  p->stamp = i;
  return 1;
}

bool peer_pause_store(peer_t *p){
  char path[PATH_BUF];

  FMT_PATH(path,DIR_PEER"/%s/"FILE_PAUSE,p->handle[0]);
  if(!bool_store(path,p->paused)){
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool peer_pause_load(peer_t *p){
  char path[PATH_BUF];

  FMT_PATH(path,DIR_PEER"/%s/"FILE_PAUSE,p->handle[0]);
  if(!file_exists(path)){
    return bool_store(path,p->paused);
  }else{
    return bool_load(path,&p->paused);
  }
}

bool peer_banner_store(peer_t *p){
  char path[PATH_BUF];

  FMT_PATH(path,DIR_PEER"/%s/"FILE_BANNER,p->handle[0]);
  if(!str_store(path,p->banner)){
    disk_unsync();
    return 0;
  }
  return 1;
}

static bool peer_banner_load(peer_t *p){
  char  path [PATH_BUF];
  char  buf  [BANNER_LEN+WHITE_LEN+1+1]; //banner\n\0
  char *b;

  FMT_PATH(path,DIR_PEER"/%s/"FILE_BANNER,p->handle[0]);
  if(!file_exists(path)){
    return file_touch(path);
  }
  if(!str_load(path,buf,sizeof(buf))){
    return 0;
  }

  b = str_trim(buf);
  if(strlen(b) > BANNER_LEN){
    log_err("banner too long (max %zu)",BANNER_LEN);
    return 0;
  }

  strzcpy(p->banner,b,sizeof(p->banner));
  return 1;
}

static bool peer_chainmess_load(peer_t *p){
  char path[PATH_BUF];

  if(!mess_load(&p->mess))        return 0;
  if(!chain_load(&p->dchain))     return 0;
  if(!chain_load(&p->bchain))     return 0;
  if(!ht_load(&p->dhead))         return 0;
  if(!ht_load(&p->dtail))         return 0;
  if(!ht_load(&p->bhead))         return 0;
  if(!ht_load(&p->btail))         return 0;

  FMT_PATH(path,DIR_PEER"/%s/"DIR_DCHAIN"/"FILE_SELF,p->handle[0]);
  if(file_exists(path)){
    if(!hash_load(path,p->self)) return 0;
  }else{
    hash_zero(p->self);
    if(!hash_store(path,p->self)) return 0;
  }

  return 1;
}

bool peer_loadpeer(peer_t *p){
  char        path[PATH_BUF];
  char const *handle = p->handle[0];

  FMT_PATH(path,DIR_PEER"/%s",handle);
  if(!mkdir_p(path))            return 0;

  if(!peer_alias_load(p))       return 0;
  if(!peer_key_load(p))         return 0;
  if(!peer_addr_load(p))        return 0;
  if(!peer_stamp_load(p))       return 0;
  if(!peer_pause_load(p))       return 0;
  if(!peer_banner_load(p))      return 0;
  if(!peer_chainmess_load(p))   return 0;

  return 1;
}

bool peer_load(void){
  char path[PATH_BUF];

  FMT_PATH(path,"%s",DIR_PEER);
  if(!mkdir_p(path)) return 0;

  //TODO: put shit in io.c
  //TODO: order??
  DIR            *d;
  struct dirent  *de;
  if((d = opendir(path)) == 0){
    log_errno("opendir");
    log_err("failed to list dir '%s'",path);
    return 0;
  }

  while(de = readdir(d)){
    peer_t *p;

    if(strcmp(de->d_name,"." ) == 0) continue;
    if(strcmp(de->d_name,"..") == 0) continue;

    p = peer_add_internal(de->d_name,1);
    if(!p) return 0;
  }
  closedir(d);

  return 1;
}

