#include "pest.h"
#include "irc.h"
#include "ircd.h"
#include "addr.h"
#include "packet.h"
#include "peer.h"
#include "chain.h"
#include "crypto.h"
#include "wipe.h"
#include "net.h"
#include "io.h"
#include "log.h"
#include "time.h"
#include "util.h"
#include "string.h"
#include "config.h"
#include "state.h"

bool pest_init(void){
  char path[PATH_BUF];

  if(!knob_load()) return 0;
  dedup_load();

  FMT_PATH(path,"%s",DIR_MESS);
  mess_init(&mess);
  mess_path(&mess,path);
  if(!mess_load(&mess)){
    irclog_err("failed to load mess: '%s'",path);
    return 0;
  }

  FMT_PATH(path,"%s",DIR_CHAIN);
  if(!mkdir_p(path)) return 0;

  FMT_PATH(path,DIR_CHAIN"/%s",FILE_SELF);
  if(file_exists(path)){
    if(!hash_load(path,self)) return 0;
  }else{
    hash_zero(self);
    if(!hash_store(path,self)) return 0;
  }

  FMT_PATH(path,DIR_CHAIN"/%s",FILE_NET);
  if(file_exists(path)){
    if(!hash_load(path,net)) return 0;
  }else{
    hash_zero(self);
    if(!hash_store(path,net)) return 0;
  }

  console_stamp(0,mono_sec());

  return 1;
}

static void peer_check_active(uint64_t time){
  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p       = peer+x;
    bool    active  = p->active;

    p->active = time < p->pak_stamp+irctime;
    if(!active && p->active){
      irc_join(p->handle[0]);
    }
    if(active && !p->active){
      irc_part(p->handle[0]);
    }
  }
}

void pest_tick(uint64_t time,uint64_t ntime,bool active){
  if(!active || !udp_bound()){
    udp_recv(0,0,pest_port,0,0);
    if(!active) return;
  }

  dedup_clean         (time);
  hearsay_process     (ntime);
  pending_process     (ntime);
  getdata_process     (ntime);
  getdata_retry       (time);
  rekey_clean         (time);
  rekey_broadcast     (time);
  banner_broadcast    (time);
  addrcast_broadcast  (time);
  ignore_broadcast    (ntime);
  peer_check_active   (time);

  for(size_t x = 0;x < PEST_RECV_MAX;x++){
    if(!pest_recv(time)) break;
  }
}

static void pestop_wot(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
    irc_out("usage: WOT [handle]");
    return;
  }

  if(c->argc > 0){
    char   *handle  = c->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    peer_print(p);
  }else{
    peer_wot_print();
  }
}

static void pestop_peer(pestcmd_t const *c){
  char    host[IRC_HOST_BUF];
  peer_t *p;

  if(c->argc != 1 || c->trail){
    irc_out("usage: PEER handle");
    return;
  }

  char *handle = c->argv[0];
  p = peer_add(handle);
  if(!p) return;
  fmt_peerhost(host,sizeof(host),p->handle[0]);
}

static void pestop_unpeer(pestcmd_t const *c){
  char host[IRC_HOST_BUF];

  if(c->argc != 1 || c->trail){
    irc_out("usage: UNPEER handle");
    return;
  }

  char   *handle  = c->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  fmt_peerhost(host,sizeof(host),p->handle[0]);
  if(p->active){
    ircd_sendf(":%s PART %s",host,irc_chan);
  }
  peer_del(p);
}

static void pestop_aka(pestcmd_t const *c){
  if(c->argc != 2 || c->trail){
    irc_out("usage: AKA handle alias");
    return;
  }

  char   *handle  = c->argv[0];
  char   *alias   = c->argv[1];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  if(!peer_alias_add(p,alias)) return;
}

static void pestop_unaka(pestcmd_t const *c){
  char host[IRC_HOST_BUF];

  if(c->argc != 1 || c->trail){
    irc_out("usage: UNAKA handle");
    return;
  }

  char   *handle  = c->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;

  fmt_peerhost(host,sizeof(host),p->handle[0]);
  unsigned r = peer_alias_del(p,handle);
  if(!r) return;
  if(r == 2){ //TODO: this is disgusting but works
    ircd_sendf(":%s NICK :%s",host,p->handle[0]);
  }
}

static void pestop_pause(pestcmd_t const *c){
  if(c->argc != 1 || c->trail){
    irc_out("usage: PAUSE handle");
    return;
  }

  char   *handle  = c->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  peer_pause(p,1);
}

static void pestop_unpause(pestcmd_t const *c){
  if(c->argc != 1 || c->trail){
    irc_out("usage: UNPAUSE handle");
    return;
  }

  char   *handle  = c->argv[0];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;
  peer_pause(p,0);
}

static void pestop_key(pestcmd_t const *c){
  if(c->argc != 2 || c->trail){
    irc_out("usage: KEY handle key");
    return;
  }

  uint8_t key[KEY_SIZE];
  char   *handle  = c->argv[0];
  char   *b64_key = c->argv[1];
  peer_t *p       = peer_lookup(handle);
  if(!p) return;

  if(!key_from_b64(key,b64_key)){
    irc_err("invalid key: '%s'",b64_key);
    key_zero(key);
    return;
  }

  peer_key_add(p,key,0);
  key_zero(key);
}

static void pestop_unkey(pestcmd_t const *c){
  uint8_t key[KEY_SIZE];

  if(c->argc != 1 || c->trail){
    irc_out("usage: UNKEY key");
    return;
  }

  char *b64_key = c->argv[0];
  if(!key_from_b64(key,b64_key)){
    irc_err("invalid key: '%s'",b64_key);
    key_zero(key);
    return;
  }

  peer_key_del(key,0);
  key_zero(key);
}

static void pestop_genkey(pestcmd_t const *c){
  uint8_t key     [KEY_SIZE];
  char    b64_key [B64_KEY_BUF];

  if(c->argc != 0 || c->trail){
    irc_out("usage: GENKEY");
    return;
  }

  if(!key_gen(key)){
    return;
  }

  b64_from_key(b64_key,sizeof(b64_key),key);
  irc_out("key generated from '%s': %s",RNG_SLOW_PATH,b64_key);
  key_zero(key);
  BUF_ZERO(b64_key);
}

static void pestop_rekey(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
    irc_out("usage: REKEY [handle]");
    return;
  }

  if(c->argc == 0){
    rekey_initiate(0,1);
  }else{
    char   *handle  = c->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    rekey_initiate(p,1);
  }
}

static void pestop_rktog(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
  usage:
    irc_out("usage: RKTOG ['ENABLE'/'DISABLE']");
    return;
  }

  if(c->argc == 0){
    knob_display(FILE_RKRECV,0);
  }else if(strcmpci(c->argv[0],"ENABLE") == 0){
    knob_set(FILE_RKRECV,"1");
  }else if(strcmpci(c->argv[0],"DISABLE") == 0){
    knob_set(FILE_RKRECV,"0");
  }else{
    goto usage;
  }
}

static void pestop_gag(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
    irc_out("usage: GAG [handle]");
    return;
  }

  if(c->argc == 0){
    killfile_print();
  }else{
    char *handle = c->argv[0];
    killfile_add(handle);
  }
}

static void pestop_ungag(pestcmd_t const *c){
  if(c->argc != 1 || c->trail){
    irc_out("usage: UNGAG handle");
    return;
  }

  char *handle = c->argv[0];
  killfile_del(handle);
}

static void pestop_at(pestcmd_t const *c){
  if(c->argc > 2 || c->trail){
    irc_out("usage: AT [handle [ip:port]]");
    return;
  }

  if(c->argc == 0){
    peer_at_print();
  }else if(c->argc == 1){
    char   *handle  = c->argv[0];
    peer_t *p       = peer_lookup(handle);
    if(!p) return;
    if(addr_null(&p->addr)){
      irc_out("no address stored for peer '%s'",handle);
      return;
    }
    peer_addr_print(p);
  }else{
    char   *handle  = c->argv[0];
    char   *at      = c->argv[1];
    peer_t *p       = peer_lookup(handle);
    addr_t  addr;

    if(!p) return;
    if(!addr_from_at(&addr,at)){
      irc_err("invalid address (format is ip:port): '%s'",at);
      return;
    }

    peer_addr_set(p,&addr,1);
  }
}

static void pestop_knob(pestcmd_t const *c){
  if((!c->trail && c->argc  > 2) ||
     ( c->trail && c->argc != 1)){
    irc_out("usage: KNOB [knob [value]]");
    return;
  }

  if(c->argc == 0){
    knob_display(0,0);
  }else if(c->trail){
    char *knob  = c->argv[0];
    char *value = c->trail;
    knob_set(knob,value);
  }else if(c->argc == 1){
    char *knob  = c->argv[0];
    knob_display(knob,0);
  }else{
    char *knob  = c->argv[0];
    char *value = c->argv[1];
    knob_set(knob,value);
  }
}

static void pestop_knobinfo(pestcmd_t const *c){
  if(c->argc != 1 || c->trail){
    irc_out("usage: KNOBINFO knob");
    return;
  }

  char *knob = c->argv[0];
  knob_display(knob,1);
}

static void pestop_cut(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
    irc_out("usage: CUT [cutoff]");
    return;
  }

  if(c->argc == 0){
    knob_display(FILE_CUTOFF,0);
  }else{
    knob_set(FILE_CUTOFF,c->argv[0]);
  }
}

static bool resolve(char const *handle,
                    char const *chname,
                    chain_t    *c,
                    ht_t       *head,
                    ht_t       *tail){

  hash_t *l;
  hash_t  h;
  hash_t  t;

  if(ht_forks (head,tail) == 0 &&
     ht_breaks(head,tail) == 0){
    return 0;
  }

  l = ht_last(head);
  if(!l){
    irc_err("make sure you have seen at least one "
            "message from peer before resolve fork/break");
    return 1;
  }
  hash_copy(h,*l);

  if(chain_resolve(c,h,t)){
    ht_clear(head,h);
    ht_clear(tail,t);
    irclog_out("peer '%s' %s fork/break resolved",
               handle,chname);
  }else{
    irclog_err("failed to resolve peer '%s' %s fork/break",
               handle,chname);
  }

  return 1;
}

static void pest_resolve(char const *handle){
  size_t  n = 0;
  peer_t *p = peer_lookup(handle);
  if(!p) return;

  if(resolve(p->handle[0],"broadcast",
             &p->bchain,&p->bhead,&p->btail)){
    n++;
  }

  if(resolve(p->handle[0],"direct message",
             &p->dchain,&p->dhead,&p->dtail)){
    n++;
  }

  if(!n){
    irc_err("cannot resolve, peer '%s' chain is not forked/broken",
            p->handle[0]);
  }
}

static void pestop_resolve(pestcmd_t const *c){
  if(c->argc != 1 || c->trail){
    irc_out("usage: RESOLVE handle");
    return;
  }

  char *handle = c->argv[0];
  pest_resolve(handle);
}

static void pestop_slave(pestcmd_t const *c){
  if(c->argc > 1 || c->trail){
    irc_out("usage: SLAVE [handle]");
    return;
  }

  if(!c->argc){
    if(*master){
      irc_out("slave mode is enabled (master = '%s')",master);
    }else{
      irc_out("slave mode is disabled");
    }
    return;
  }

  if(*master){
    irc_err("slave mode already enabled");
    return;
  }

  char *handle = c->argv[0];
  knob_set(FILE_MASTER,handle);
  irc_out("slave mode enabled");
}

static void pestop_unslave(pestcmd_t const *c){
  if(c->argc != 0 || c->trail){
    irc_out("usage: UNSLAVE");
    return;
  }

  if(!*master){
    irc_err("slave mode already disabled");
    return;
  }

  knob_set(FILE_MASTER,"");
  irc_out("slave mode disabled");
}

static bool pest_achtung(char const *msg){
  uint16_t peer_idx[peer_cnt];

  if(!index_shuffle(peer_idx,peer_cnt)){
    irclog_err("could not shuffle index, achtung not sent");
    return 0;
  }

  for(size_t x = 0;x < peer_cnt;x++){
    peer_t *p = peer+peer_idx[x];
    if(!peer_cansend(p,0)) continue;
    send_ustr(CMD_DIRECT,msg,p);
  }

  return 1;
}

static void pestop_achtung(pestcmd_t const *c){
  if(!c->trail){
    irc_out("usage: ACHTUNG message");
    return;
  }

  char *msg = c->trail;
  pest_achtung(msg);
}

static void pestop_scram(pestcmd_t const *c){
  char     *pwd;
  uint8_t   pass[SHA512_SIZE];

  if(!c->trail){
    irc_out("usage: SCRAM scrampass");
    return;
  }

  pwd = c->trail;
  sha512(pass,(uint8_t *)pwd,strlen(pwd));
  if(memcmp(pass,scram_pass,SHA512_SIZE) != 0){
    irc_err("SCRAM password mismatch");
    return;
  }

  scram_push();
}

#define PEST_OP(N,A)  { #N, A, pestop_##N },

typedef void (*pestcmd_f)(pestcmd_t const *);

typedef struct pestop{
  char const       *cmd;
  size_t            arg_max;
  pestcmd_f         func;
}pestop_t;

static pestop_t const pest_op[] = {
  PEST_OP (wot,         1)
  PEST_OP (peer,        1)
  PEST_OP (unpeer,      1)
  PEST_OP (aka,         2)
  PEST_OP (unaka,       1)
  PEST_OP (pause,       1)
  PEST_OP (unpause,     1)
  PEST_OP (key,         2)
  PEST_OP (unkey,       1)
  PEST_OP (genkey,      0)
  PEST_OP (rekey,       1)
  PEST_OP (rktog,       1)
  PEST_OP (gag,         1)
  PEST_OP (ungag,       1)
  PEST_OP (at,          2)
  PEST_OP (knob,        1)
  PEST_OP (knobinfo,    1)
  PEST_OP (cut,         1)
  PEST_OP (resolve,     1)
  PEST_OP (slave,       1)
  PEST_OP (unslave,     1)
  PEST_OP (achtung,     0)
  PEST_OP (scram,       0)
};

static pestop_t const *get_op(char const *cmd){
  char cmd_buf[PEST_CMD_BUF];

  for(size_t x = 0;x < ARR_CNT(pest_op);x++){
    pestop_t const *op = pest_op+x;
    if(strncmpci(cmd,op->cmd,IRCD_CMD_BUF) != 0) continue;
    return op;
  }

  strzcpy(cmd_buf,cmd,sizeof(cmd_buf));
  str_toupper(cmd_buf);
  irc_err("unknown/unhandled pest command received: %s",cmd_buf);
  return 0;
}

static pestop_t const *parse_cmd(pestcmd_t *c,char *cmd){
  pestop_t const *op;
  char           *p;
  char           *l;

  c->cmd    = cmd;
  c->argc   = 0;
  c->trail  = 0;

  p = scan_white(cmd);
  if(!*p){
    return get_op(c->cmd);
  }
  *p++ = 0;

  op = get_op(c->cmd);
  if(!op) return 0;

  p = scan_nowhite(p);
  l = p;
  while(*p){
    p = scan_white(p);
    if(c->argc < op->arg_max){
      c->argv[c->argc] = l;
      c->argc++;
    }else{
      c->trail = l;
      break;
    }
    if(!*p) break;
    *p++  = 0;
    p     = scan_nowhite(p);
    l     = p;
  }

  return op;
}

static bool exec_cmd(char const *cmd){
  pestcmd_t         c;
  pestop_t const   *op;
  char              cmd_buf[IRC_MSG_BUF];

  strzcpy(cmd_buf,cmd,sizeof(cmd_buf));
  op = parse_cmd(&c,cmd_buf);
  if(!op) return 0;
  op->func(&c);
  return 1;
}

static char const *msg_cmdp(char const *msg){
  char const *f = scan_nowhite(msg);
  return f[0] == '%' && f[1] != '%' ? ++f : 0;
}

static char const *msg_esc(char const *msg){
  return msg[0] == '%' && msg[1] == '%' ? msg+1 : msg;
}

bool pest_broadcast(char const *msg){
  char const *cmd = msg_cmdp(msg);
  if(cmd){
    return exec_cmd(cmd);
  }else{
    return send_ustr(CMD_BROADCAST,msg_esc(msg),0);
  }
}

bool pest_direct(char const *handle,char const *msg){
  char const *cmd = msg_cmdp(msg);
  bool        r   = 1;

  irc_log_chan = handle;
  if(cmd){
    r = exec_cmd(cmd);
  }else{
    peer_t *p = peer_lookup(handle);
    if(!p){
      ircd_sendf(":%s 401 %s %s :No such nick",
                 handle,self_handle,handle);
      return 0;
    }
    if(!peer_cansend(p,1)) return 0;
    r = send_ustr(CMD_DIRECT,msg_esc(msg),p);
  }

  irc_log_chan = 0;
  return r;
}

