#include "pest.h"
#include "peer.h"
#include "crypto.h"
#include "packet.h"
#include "log.h"
#include "time.h"
#include "util.h"
#include "state.h"

//TODO: retry on no response

#define STATE_INACTIVE WAIT_OFFER1

enum{
  WAIT_OFFER1,
  WAIT_OFFER2,
  WAIT_SLICE1,
  WAIT_SLICE2,
  WAIT_IGNORE1,
  WAIT_IGNORE2,
  WAIT_PACKET,
};

bool rekey_active(rekey_t *rk){
  return rk->state != STATE_INACTIVE;
}

bool rekey_keyexists(rekey_t *rk){
  return rk->state == WAIT_IGNORE1 ||
         rk->state == WAIT_IGNORE2 ||
         rk->state == WAIT_PACKET;
}

static void rekey_abort(peer_t *p,char const *reason){
  rekey_t  *rk  = &p->rekey;
  log_f     err = rk->log ? irclog_err : log_err;

  err("rekeying with peer '%s' aborted: %s",p->handle[0],reason); 
  rekey_init(rk);
}

static bool keyoffer_send(peer_t *p){
  rekey_t  *rk  = &p->rekey;

  if(!key_gen(rk->slice_self)){
    rekey_abort(p,"could not generate key slice");
    return 0;
  }
  sha512(rk->offer_self,rk->slice_self,sizeof(rk->slice_self));

  if(!send_text(CMD_KEYOFFER,
                mono_sec(),
                rk->offer_self,
                sizeof(rk->offer_self),
                p,
                0)){
    rekey_abort(p,"could not send data");
    return 0;
  }

  return 1;
}

static bool keyslice_send(peer_t *p){
  rekey_t    *rk      = &p->rekey;

  if(!send_text(CMD_KEYSLICE,
                mono_sec(),
                rk->slice_self,
                sizeof(rk->slice_self),
                p,
                0)){
    rekey_abort(p,"could not send data");
    return 0;
  }

  return 1;
}

static bool keyoffer_recv(peer_t *p,uint8_t const *offer){
  rekey_t *rk = &p->rekey;

  memcpy(rk->offer_peer,offer,sizeof(rk->offer_peer));

  if(memcmp(rk->offer_peer,
            rk->offer_self,
            sizeof(rk->offer_peer)) == 0){
    rekey_abort(p,"received key offer was "
                  "identical to key offer sent");
    return 0;
  }

  return 1;
}

static bool keyslice_recv(peer_t *p,uint8_t const *slice){
  uint8_t   hash[SHA512_SIZE];
  rekey_t  *rk = &p->rekey;
  pkey_t   *pk;

  key_copy(rk->slice_peer,slice);
  sha512(hash,rk->slice_peer,sizeof(rk->slice_peer));

  if(memcmp(hash,rk->offer_peer,sizeof(hash)) != 0){
    BUF_ZERO(hash);
    rekey_abort(p,"received key slice did not match key offer");
    return 0;
  }
  BUF_ZERO(hash);

  if(entropy_missing(rk->slice_peer,sizeof(rk->slice_peer))){
    rekey_abort(p,"entropy obviously missing "
                  "from received key slice");
    return 0;
  }

  pk = peer_key_recent(p);
  assert(pk);

  key_copy(rk->key_old,pk->key.full);
  key_xor(rk->key_new,rk->key_old,rk->slice_self);
  key_xor(rk->key_new,rk->key_new,rk->slice_peer);

  if(!peer_key_add(p,rk->key_new,1)){
    rekey_abort(p,"could not add new key");
    return 0;
  }

  return 1;
}

static pkey_t *rekey_pkey(peer_t *p){
  rekey_t  *rk = &p->rekey;
  pkey_t   *pk = peer_lookup_pkey(p,rk->key_new);
  if(!pk){
    rekey_abort(p,"new key has been deleted");
    return 0;
  }
  return pk;
}

static bool rkignore_send(peer_t *p){
  pkey_t *pk;

  pk = rekey_pkey(p);
  if(!pk) return 0;

  return ignore_send(p,pk);
}

static bool rkignore_recv(peer_t *p,pkey_t *pk,uint64_t time){
  rekey_t *rk = &p->rekey;
  return key_equal(pk->key.full,rk->key_new);
}

static bool packet_recv(peer_t *p,pkey_t *pk){
  rekey_t *rk = &p->rekey;

  if(!key_equal(pk->key.full,rk->key_new)){
    //TODO: delete key here or vat?
    return 0;
  }

  rk->pak_cnt++;
  if(rk->pak_cnt >= REKEY_PACKET_CNT){
    peer_key_del(rk->key_old,1);
    rekey_init(rk);
  }

  return 1;
}

static void rekey_complete(peer_t *p){
  rekey_t *rk = &p->rekey;
  rk->state   = WAIT_PACKET;
  rk->pak_cnt = 0;
  irclog_out("rekeying with peer '%s' successful",
             p->handle[0]);   
}

static bool initiate(peer_t *p,bool log){
  rekey_t  *rk  = &p->rekey;
  log_f     err = log ? irclog_err : log_err;

  if(!peer_cansend(p,log)) return 0;

  if(rekey_active(rk)){
    err("rekeying with peer '%s' already in progress",p->handle[0]);
    return 0;
  }

  if(!keyoffer_send(p)) return 0;

  rk->state   = WAIT_OFFER2;
  rk->log     = log;
  rk->time    = mono_sec();
  rk->pak_cnt = 0;

  return 1;
}

void rekey_initiate(peer_t *p,bool log){
  log_f out = log ? irclog_out : log_out;

  if(p){
    if(initiate(p,log)){
      out("rekeying with peer '%s' initiated",p->handle[0]);
    }
  }else{
    uint16_t peer_idx[peer_cnt];

    if(!index_shuffle(peer_idx,peer_cnt)){
      irclog_err("could not shuffle index, "
                 "rekeying cannot be initiated");
      return;
    }

    for(size_t x = 0;x < peer_cnt;x++){
      peer_t *q = peer+peer_idx[x];
      if(!peer_cansend(q,0))      continue;
      if(rekey_active(&q->rekey)) continue;
      initiate(q,log);
    }

    out("rekeying with all peers initiated");
  }
}

void rekey_init(rekey_t *rk){
  rk->state     = STATE_INACTIVE;
  rk->log       = 0;
  rk->time      = 0;
  rk->pak_cnt   = 0;

  BUF_ZERO(rk->offer_self);
  BUF_ZERO(rk->offer_peer);
  BUF_ZERO(rk->slice_self);
  BUF_ZERO(rk->slice_peer);
  BUF_ZERO(rk->key_old);
  BUF_ZERO(rk->key_new);
}

bool rekey_handle_offer(peer_t *p,uint8_t const *offer){
  rekey_t *rk = &p->rekey;
  if(!peer_cansend(p,rk->log)) return 0;

  switch(rk->state){
  case WAIT_OFFER1:
    if(!rkrecv)                   return 1;
    if(!peer_cansend(p,rk->log))  return 0;
    if(!keyoffer_recv(p,offer))   return 0;
    if(!keyoffer_send(p))         return 0;
    rk->state   = WAIT_SLICE1;
    rk->log     = 0;
    rk->time    = mono_sec();
    rk->pak_cnt = 0;
    return 1;

  case WAIT_OFFER2:
    if(!peer_cansend(p,rk->log))  return 0;
    if(!keyoffer_recv(p,offer))   return 0;
    if(!keyslice_send(p))         return 0;
    rk->state = WAIT_SLICE2;
    return 1;

  default:
    rekey_abort(p,"invalid sequence");
    return 0;
  }
}

bool rekey_handle_slice(peer_t *p,uint8_t const *slice){
  rekey_t *rk = &p->rekey;

  switch(rk->state){
  case STATE_INACTIVE:
    return 0;

  case WAIT_SLICE1:
    if(!peer_cansend(p,rk->log))  return 0;
    if(!keyslice_recv(p,slice))   return 0;
    if(!keyslice_send(p))         return 0;
    rk->state = WAIT_IGNORE1;
    return 1;

  case WAIT_SLICE2:
    if(!peer_cansend(p,rk->log))  return 0;
    if(!keyslice_recv(p,slice))   return 0;
    if(!rkignore_send(p))         return 0;
    rk->state = WAIT_IGNORE2;
    return 1;

  default:
    rekey_abort(p,"invalid sequence");
    return 0;
  }
}

bool rekey_handle_ignore(peer_t *p,pkey_t *pk,uint64_t time){
  rekey_t *rk = &p->rekey;

  switch(rk->state){
  case STATE_INACTIVE:
    return 0;

  case WAIT_IGNORE1:
    if(!peer_cansend(p,rk->log))  return 0;
    if(!rkignore_recv(p,pk,time)) return 0;
    if(!rkignore_send(p))         return 0;
    rekey_complete(p);
    return 1;

  case WAIT_IGNORE2:
    if(!peer_cansend(p,rk->log))  return 0;
    if(!rkignore_recv(p,pk,time)) return 0;
    rekey_complete(p);
    return 1;

  default:
    return 0;
  }
}

bool rekey_handle_packet(peer_t *p,pkey_t *pk){
  rekey_t *rk = &p->rekey;

  switch(rk->state){
  case STATE_INACTIVE:
    return 0;

  case WAIT_PACKET:
    if(!peer_cansend(p,rk->log)) return 0;
    if(!packet_recv(p,pk))       return 0;
    return 1;

  default:
    return 0;
  }
}

void rekey_broadcast(uint64_t time){
  if(time < rekey_last+rkint*SEC_HOUR) return;
  if(rksend && rekey_last != 0){
    rekey_initiate(0,0);
  }
  rekey_last = time;
}

void rekey_clean(uint64_t time){
  for(size_t x = 0;x < peer_cnt;x++){
    peer_t   *p   = peer+x;
    rekey_t  *rk  = &p->rekey;
    log_f     out = rk->log ? irclog_out : log_out;

    if(!rekey_active(rk))         continue;
    if(time < rk->time+rktime)    continue;
    if(rk->state == WAIT_PACKET)  continue;

    if(rekey_keyexists(rk)){
      peer_key_del(rk->key_new,1);
    }
    rekey_init(rk);
    out("rekeying with peer '%s' timed out",p->handle[0]);
  }
}

