#include "packet.h"
#include "pest.h"
#include "peer.h"
#include "chain.h"
#include "log.h"
#include "net.h"
#include "crypto.h"
#include "io.h"
#include "serial.h"
#include "utf8.h"
#include "time.h"
#include "util.h"
#include "type.h"
#include "state.h"
#include "config.h"

static bool red_send(uint8_t *r,
                     peer_t  *p,
                     pkey_t  *pk){

  uint8_t b[BLACK_SIZE];

  if(!peer_cansend(p,0)){
    return 1;
  }

  if(!rng_read(rng_fast_fd,r+NONCE_OFF,NONCE_SIZE)){
    return 0;
  }

  if(pk){
#ifndef NDEBUG
    peer_t *q;
    assert(pkey_lookup_noerr(pk->key.full,&q));
    assert(q == p);
#endif
  }else{
    pk = peer_key_recent(p);
    assert(pk);
  }

  black_ciph(b,r,pk->ciph_serp);
  black_sign(b,pk->key.sub.sign);

  if(udp_send(p->addr.ip,p->addr.port,b,sizeof(b)) != sizeof(b)){
    irclog_err("failed to send UDP packet to peer '%s'",p->handle[0]);
    return 0;
  }

  return 1;
}

static bool red_send_many(uint8_t      *r,
                          hash_t const  hash,
                          peer_t       *chan,
                          pkey_t       *pk){

  peer_t   *p       = chan ? chan : peer;
  size_t    p_cnt   = chan ? 1    : peer_cnt;
  uint16_t  peer_idx[p_cnt];
  bool      ret;
  uint8_t   command;

  assert(!pk || chan);

  if(p_cnt == 1){
    peer_idx[0] = 0;
  }else if(!index_shuffle(peer_idx,p_cnt)){
    irclog_err("could not shuffle index, "
               "message cannot be sent");
    return 0;
  }

  read_u8(r+COMMAND_OFF,&command);

  ret = 1;
  for(size_t x = 0;x < p_cnt;x++){
    peer_t *q = p+peer_idx[x];
    if(!red_send(r,q,pk)) ret = 0;
#ifndef LOG_COLLAPSE
    log_packet(hash,command,q,&q->addr,1,0,0);
#endif
  }

#ifdef LOG_COLLAPSE
  log_packet(hash,command,chan,chan ? &chan->addr : 0,1,0,0);
#endif

  return ret;
}

bool red_rebroadcast(uint8_t *r,hash_t const hash,uint8_t bounces){
  if(bounces >= BOUNCES_MAX) return 1;
  bounces++;
  write_u8(r+BOUNCES_OFF,&bounces);
  return red_send_many(r,hash,0,0);
}

bool send_message(uint8_t   command,
                  uint8_t  *r,
                  hash_t    hash,
                  peer_t   *chan,
                  pkey_t   *pk){

  uint8_t     bounces;
  uint8_t     version;
  uint8_t     reserved;

  assert(!pk || chan);

  bounces   = 0;
  version   = PEST_VERSION;
  reserved  = 0;

  write_u8  (r+BOUNCES_OFF,   &bounces);
  write_u8  (r+VERSION_OFF,   &version);
  write_u8  (r+RESERVED_OFF,  &reserved);
  write_u8  (r+COMMAND_OFF,   &command);

  return red_send_many(r,hash,chan,pk);
}

static void set_hash(hash_t        *dst,
                     hash_t const   src,
                     peer_t        *p,
                     char const    *file){

  char path[PATH_BUF];
  if(p){
    FMT_PATH(path,DIR_PEER"/%s/"DIR_DCHAIN"/%s",p->handle[0],file);
  }else{
    FMT_PATH(path,DIR_CHAIN"/%s",file);
  }
  hash_copy(*dst,src);
  if(!hash_store(path,*dst)){
    disk_unsync();
  }
}

bool send_text(uint8_t        command,
               uint64_t       stamp,
               uint8_t const *text,
               size_t         text_size,
               peer_t        *chan,
               pkey_t        *pk){

  uint8_t     r[RED_SIZE];
  uint8_t    *m;
  hash_t      hash;
  bool        ret;

  m = r+MESSAGE_OFF;
  write_u64(m+TIMESTAMP_OFF,&stamp);

  strncpy((char *)m+SPEAKER_OFF,self_handle,SPEAKER_SIZE);

  assert(text_size <= TEXT_SIZE);
  memcpy(m+TEXT_OFF,text,text_size);
  memset(m+TEXT_OFF+text_size,0x00,TEXT_SIZE-text_size);

  switch(command){
  case CMD_BROADCAST:
    hash_copy(m+SELFCHAIN_OFF,self);
    hash_copy(m+NETCHAIN_OFF,net);
    break;

  case CMD_DIRECT:
    assert(chan);
    hash_copy(m+SELFCHAIN_OFF,chan->self);
    hash_zero(m+NETCHAIN_OFF);
    break;

  default:
    hash_zero(m+SELFCHAIN_OFF);
    hash_zero(m+NETCHAIN_OFF);
  }

  hash_data(hash,m,MESSAGE_SIZE);

  switch(command){
  case CMD_BROADCAST:
    set_hash(&self,hash,0,FILE_SELF);
    set_hash(&net, hash,0,FILE_NET);
    mess_push(&mess,hash,m);
    dedup_push(hash,stamp);
    break;

  case CMD_DIRECT:
    assert(chan);
    set_hash(&chan->self,hash,chan,FILE_SELF);
    mess_push(&chan->mess,hash,m);
    dedup_push(hash,stamp);
    break;
  }

  ret = send_message(command,r,hash,chan,pk);
  BUF_ZERO(r);
  return ret;
}

bool send_ustr(uint8_t cmd,char const *ustr,peer_t *chan){
  char const *ustr_orig;
  uint64_t    stamp;
  size_t      ustr_len;
  bool        ret;

  assert(utf8_valid(ustr));

  ret       = 1;
  ustr_orig = ustr;
  ustr_len  = strlen(ustr);
  stamp     = mono_sec();

  do{
    size_t text_len = ustr_len;
    if(text_len > TEXT_SIZE){
      text_len = utf8_cram_unsafe(ustr,TEXT_SIZE);
    }

    if(!send_text(cmd,
                  stamp,
                  (uint8_t const *)ustr,
                  text_len,
                  chan,
                  0)){
      ret = 0;
    }

    ustr     += text_len;
    ustr_len -= text_len;
  }while(ustr_len > 0);

  if(!ret){
    irclog_err("pest message (%.10s...) "
               "was not sent to all peers",
               ustr_orig);
  }

  console_stamp(chan,stamp);

  return ret;
}

