#include "crypto.h"
#include "serial.h"
#include <string.h>

#define SHA256_BLOCK_SIZE     (512/8)

#define BYTE_BITS             8
#define WORD_SIZE             4
#define BLOCK_BYTES           SHA256_BLOCK_SIZE
#define BLOCK_BITS            (BLOCK_BYTES*8)
#define BLOCK_WORDS           (BLOCK_BYTES/WORD_SIZE)
#define LEN_BITS             64
#define LEN_BYTES             (LEN_BITS/BYTE_BITS)
#define ROUND_N              64
#define STATE_N               (SHA256_SIZE/WORD_SIZE)
#define OUT256_N              STATE_N

static uint32_t const h_init[STATE_N] = {
  0x6A09E667,0xBB67AE85,0x3C6EF372,0xA54FF53A,
  0x510E527F,0x9B05688C,0x1F83D9AB,0x5BE0CD19,
};

static uint32_t const k[ROUND_N] = {
  0x428A2F98,0x71374491,0xB5C0FBCF,0xE9B5DBA5,
  0x3956C25B,0x59F111F1,0x923F82A4,0xAB1C5ED5,
  0xD807AA98,0x12835B01,0x243185BE,0x550C7DC3,
  0x72BE5D74,0x80DEB1FE,0x9BDC06A7,0xC19BF174,
  0xE49B69C1,0xEFBE4786,0x0FC19DC6,0x240CA1CC,
  0x2DE92C6F,0x4A7484AA,0x5CB0A9DC,0x76F988DA,
  0x983E5152,0xA831C66D,0xB00327C8,0xBF597FC7,
  0xC6E00BF3,0xD5A79147,0x06CA6351,0x14292967,
  0x27B70A85,0x2E1B2138,0x4D2C6DFC,0x53380D13,
  0x650A7354,0x766A0ABB,0x81C2C92E,0x92722C85,
  0xA2BFE8A1,0xA81A664B,0xC24B8B70,0xC76C51A3,
  0xD192E819,0xD6990624,0xF40E3585,0x106AA070,
  0x19A4C116,0x1E376C08,0x2748774C,0x34B0BCB5,
  0x391C0CB3,0x4ED8AA4A,0x5B9CCA4F,0x682E6FF3,
  0x748F82EE,0x78A5636F,0x84C87814,0x8CC70208,
  0x90BEFFFA,0xA4506CEB,0xBEF9A3F7,0xC67178F2,
};

static inline uint32_t shr(uint32_t x,size_t n){
  return x>>n;
}

static inline uint32_t ror(uint32_t x,size_t n){
  return x>>n | x<<(32-n);
}

static inline uint32_t ch(uint32_t x,
                          uint32_t y,
                          uint32_t z){
  return ( x & y) ^
         (~x & z);
}

static inline uint32_t maj(uint32_t x,
                           uint32_t y,
                           uint32_t z){
  return (x & y) ^
         (x & z) ^
         (y & z);
}

static inline uint32_t usigma0(uint32_t x){
  return ror(x, 2) ^
         ror(x,13) ^
         ror(x,22);
}

static inline uint32_t usigma1(uint32_t x){
  return ror(x, 6) ^
         ror(x,11) ^
         ror(x,25);
}

static inline uint32_t lsigma0(uint32_t x){
  return ror(x, 7) ^
         ror(x,18) ^
         shr(x, 3);
}

static inline uint32_t lsigma1(uint32_t x){
  return ror(x,17) ^
         ror(x,19) ^
         shr(x,10);
}

static inline void expand_block(uint32_t *w,uint8_t const *block){
  for(size_t x = 0;x < BLOCK_WORDS;x++){
    uint32_t b_host;
    memcpy(&b_host,block+sizeof(uint32_t)*x,sizeof(uint32_t));
    w[x] = host32be(b_host);
  }
  for(size_t x = BLOCK_WORDS;x < ROUND_N;x++){
    w[x] = lsigma1(w[x -  2]) +
                   w[x -  7]  +
           lsigma0(w[x - 15]) +
                   w[x - 16];
  }
}

static void sha256_block(uint32_t *state,uint8_t const *block){
  uint32_t t1;
  uint32_t t2;
  uint32_t s[STATE_N];
  uint32_t w[ROUND_N];

  for(size_t x = 0;x < STATE_N;x++){
    s[x] = state[x];
  }

  expand_block(w,block);

  for(size_t x = 0;x < ROUND_N;x++){
    t1    = k[x]+w[x];
    t1   += usigma1(s[4])+ch(s[4],s[5],s[6])+s[7];
    t2    = usigma0(s[0])+maj(s[0],s[1],s[2]);
    s[7]  = s[6];
    s[6]  = s[5];
    s[5]  = s[4];
    s[4]  = s[3]+t1;
    s[3]  = s[2];
    s[2]  = s[1];
    s[1]  = s[0];
    s[0]  = t1+t2;
  }

  for(size_t x = 0;x < STATE_N;x++){
    state[x] += s[x];
  }
}

void sha256(uint8_t *out,uint8_t const *data,size_t size){
  uint64_t  size_b;
  uint64_t  spad_b;
  size_t    block_n;
  size_t    block_last_off;
  size_t    block_last_size;
  size_t    block_last_n;
  uint8_t   block_last[BLOCK_BYTES*2];
  size_t    pad80_off;
  size_t    pad00_off;
  size_t    pad00_size;
  uint64_t  len;
  size_t    len_off;
  uint32_t  state[STATE_N];

  size_b          = (uint64_t)size*BYTE_BITS;
  spad_b          = size_b+1+LEN_BITS;
  block_n         = (spad_b-1)/BLOCK_BITS+1;
  if((block_n-1)*BLOCK_BYTES > size){
    block_last_n  = 2;
  }else{
    block_last_n  = 1;
  }
  block_last_off  = (block_n-block_last_n)*BLOCK_BYTES;
  block_last_size = size-block_last_off;
  len             = host64be(size_b);
  len_off         = BLOCK_BYTES*block_last_n-LEN_BYTES;
  pad80_off       = block_last_size;
  pad00_off       = pad80_off+1;
  pad00_size      = len_off-pad00_off;

  memcpy(block_last,data+block_last_off,block_last_size);
  block_last[pad80_off] = 0x80;
  memset(block_last+pad00_off,0x00,pad00_size);
  memcpy(block_last+len_off,&len,LEN_BYTES);
  memcpy(state,h_init,sizeof(state));

  for(size_t x = 0;x < block_n-block_last_n;x++){
    sha256_block(state,data+x*BLOCK_BYTES);
  }
  for(size_t x = 0;x < block_last_n;x++){
    sha256_block(state,block_last+x*BLOCK_BYTES);
    memset_ffs(block_last+x*BLOCK_BYTES,0x00,BLOCK_BYTES);
  }

  for(size_t x = 0;x < OUT256_N;x++){
    uint32_t state_be = host32be(state[x]);
    memcpy(out+x*sizeof(uint32_t),&state_be,sizeof(uint32_t));
  }
}

